import queue
import base64
import logging
import traceback
import threading
import collections

import envi
import envi.bits as e_bits
import envi.memory as e_mem
import envi.pagelookup as e_page
import envi.codeflow as e_codeflow

import vstruct.cparse as vs_cparse
import vstruct.builder as vs_builder
import vstruct.constants as vs_const

import vivisect.const as viv_const
import vivisect.impapi as viv_impapi
import vivisect.parsers as viv_parsers
import vivisect.analysis as viv_analysis
import vivisect.codegraph as viv_codegraph

from envi.threads import firethread

from vivisect.exc import *
from vivisect.const import *

logger = logging.getLogger(__name__)

"""
Mostly this is a place to scuttle away some of the inner workings
of a workspace, so the outer facing API is a little cleaner.
"""
class VivEventCore(object):
    '''
    A class to facilitate event monitoring in the viv workspace.
    '''

    def __init__(self, vw=None, **kwargs):
        self._ve_vw = vw
        self._ve_ehand = [None for x in range(VWE_MAX)]
        self._ve_thand = [None for x in range(VTE_MAX)]
        self._ve_lock = threading.Lock()

        # Find and put handler functions into the list
        for name in dir(self):
            if name.startswith('VWE_'):
                idx = getattr(viv_const, name, None)
                self._ve_ehand[idx] = getattr(self, name)
            if name.startswith('VTE_'):
                idx = getattr(viv_const, name, None)
                self._ve_thand[idx] = getattr(self, name)

    def _ve_fireEvent(self, event, edata):
        hlist = self._ve_ehand
        if event & VTE_MASK:
            event ^= VTE_MASK
            hlist = self._ve_thand

        h = hlist[event]
        if h is not None:
            try:
                h(self._ve_vw, event, edata)
            except Exception as e:
                logger.error(traceback.format_exc())

    @firethread
    def _ve_fireListener(self):
        chanid = self._ve_vw.createEventChannel()
        try:
            etup = self._ve_vw.waitForEvent(chanid)
            while etup is not None:
                self._ve_lock.acquire()
                self._ve_lock.release()

                self._ve_fireEvent(*etup)

                etup = self._ve_vw.waitForEvent(chanid)

        finally:
            self._ve_vw.deleteEventChannel(chanid)

    def _ve_freezeEvents(self):
        self._ve_lock.acquire()

    def _ve_thawEvents(self):
        self._ve_lock.release()

vaset_xlate = {
    int:VASET_ADDRESS,
    str:VASET_STRING,
}

class VivEventDist(VivEventCore):
    '''
    Similar to an event core, but does optimized distribution
    to a set of sub eventcore objects (think GUI windows...)
    '''
    def __init__(self, vw=None, **kwargs):
        if vw is None:
            raise Exception("VivEventDist requires a vw argument")

        VivEventCore.__init__(self, vw)
        self._ve_subs = [ [] for x in range(VWE_MAX) ]
        self._ve_tsubs = [ [] for x in range(VTE_MAX) ]

        self.addEventCore(self)

        # event distributors pretty much always need a thread
        self._ve_fireListener()

    def addEventCore(self, core):
        for i in range(VWE_MAX):
            h = core._ve_ehand[i]
            if h is not None:
                self._ve_subs[i].append(h)

        for i in range(VTE_MAX):
            h = core._ve_thand[i]
            if h is not None:
                self._ve_tsubs[i].append(h)

    def delEventCore(self, core):
        for i in range(VWE_MAX):
            h = core._ve_ehand[i]
            if h is not None:
                self._ve_subs[i].remove(h)

        for i in range(VTE_MAX):
            h = core._ve_thand[i]
            if h is not None:
                self._ve_tsubs[i].remove(h)

    def _ve_fireEvent(self, event, edata):
        '''
        We don't have events of our own, we just hand them down.
        '''
        subs = self._ve_subs
        if event & VTE_MASK:
            event ^= VTE_MASK
            subs = self._ve_tsubs

        hlist = subs[event]
        for h in hlist:
            try:
                h(self._ve_vw, event, edata)
            except Exception:
                logger.error(traceback.format_exc())

        VivEventCore._ve_fireEvent(self, event, edata)


def ddict():
    return collections.defaultdict(dict)


class VivWorkspaceCore(viv_impapi.ImportApi):
    '''
    A base class that the VivWorkspace inherits from that defines a lot of the event handlers
    for things like the creation of the various location types.
    '''
    def __init__(self):
        viv_impapi.ImportApi.__init__(self)
        self.loclist = []
        self.bigend = False
        self.locmap = e_page.MapLookup()
        self.blockmap = e_page.MapLookup()
        self._mods_loaded = False
        self.parsedbin = None

        # Storage for function local symbols
        self.localsyms = ddict()

        self._call_graph = viv_codegraph.CallGraph()
        # Just in case of the GUI... :)
        self._call_graph.setMeta('bgcolor', '#000')
        self._call_graph.setMeta('nodecolor', '#00ff00')
        self._call_graph.setMeta('edgecolor', '#00802b')

        self._event_list = []
        self._event_saved = 0 # The index of the last "save" event...

        # Give ourself a structure namespace!
        self.vsbuilder = vs_builder.VStructBuilder()
        self.vsconsts  = vs_const.VSConstResolver()

        # Follow the Leader data
        self.leaders = {}       # existing "leader" sessions
        self.leaderloc = {}     # last known location for each session

    def _snapInAnalysisModules(self):
        '''
        Snap in the analysis modules which are appropriate for the
        format/architecture/platform of this workspace by calling
        '''
        if self._mods_loaded:
            return

        viv_analysis.addAnalysisModules(self)
        self._mods_loaded = True

    def _createSaveMark(self):
        '''
        Update the index of the most recent saved event to the current
        length of the event list (called after successful save)..
        '''
        self._event_saved = len(self._event_list)

    def _handleADDLOCATION(self, loc):
        lva, lsize, ltype, linfo = loc
        self.locmap.setMapLookup(lva, lsize, loc)
        self.loclist.append(loc)

        # A few special handling cases...
        if ltype == LOC_IMPORT:
            # Check if the import is registered in NoReturnApis
            if self.getMeta('NoReturnApis', {}).get(linfo.lower()):
                self.cfctx.addNoReturnAddr(lva)

    def _handleDELLOCATION(self, loc):
        # FIXME delete xrefs
        lva, lsize, ltype, linfo = loc
        self.locmap.setMapLookup(lva, lsize, None)
        self.loclist.remove(loc)

    def _handleADDSEGMENT(self, einfo):
        self.segments.append(einfo)

    def _handleADDRELOC(self, einfo):
        if len(einfo) == 4:
            fname, ptroff, rtype, data = einfo
            size = None
        else:
            fname, ptroff, rtype, data, size = einfo

        if size is None:
            size = self.psize
        imgbase = self.getFileMeta(fname, 'imagebase')
        rva = imgbase + ptroff

        self.reloc_by_va[rva] = rtype
        self.relocations.append((fname, ptroff, rtype, data, size))

        # RTYPE_BASERELOC assumes the memory is already accurate (eg. PE's unless rebased)

        if rtype in REBASE_TYPES:
            # add imgbase and offset to pointer in memory
            # 'data' arg must be 'offset' number
            ptr = imgbase + data
            if ptr != (ptr & e_bits.u_maxes[size]):
                logger.warning('Relocations calculated a bad pointer: 0x%x (imgbase: 0x%x) (relocation: %d)', ptr, imgbase, rtype)

            # writes are costly, especially on larger binaries
            if ptr != self.readMemValue(rva, size):
                with self.getAdminRights():
                    self.writeMemValue(rva, ptr, size)

        if rtype == RTYPE_BASEPTR:
            # make it like a pointer (but one that could move with each load)
            #   self.addXref(va, tova, REF_PTR)
            #   ploc = self.addLocation(va, psize, LOC_POINTER)
            #   don't follow.  handle it later, once "known code" is analyzed
            ptr, reftype, rflags = self.arch.archModifyXrefAddr(ptr, None, None)
            self._handleADDXREF((rva, ptr, REF_PTR, 0))
            self._handleADDLOCATION((rva, size, LOC_POINTER, ptr))

    def _handleDELRELOC(self, einfo):
        fname, rva, rtyp, full = einfo
        imgbase = self.getFileMeta(fname, 'imagebase')
        ptroff = rva - imgbase

        self.reloc_by_va.pop(rva, None)
        delidx = -1

        for idx, (fn, off, typ, data, size) in enumerate(self.relocations):
            if fn == fname and off == ptroff and typ == rtyp:
                delidx = idx
                break

        if delidx >= 0:
            self.relocations.pop(delidx)

        if full:
            if rtyp == RTYPE_BASEPTR:
                ptr = imgbase + data
                ptr, reftype, rflags = self.arch.archModifyXrefAddr(ptr, None, None)
                self._handleDELXREF((rva, ptr, REF_PTR, 0))
                self._handleDELLOCATION((rva, self.psize, LOC_POINTER, ptr))

    def _handleADDMODULE(self, einfo):
        logger.warning('DEPRECATED (ADDMODULE) ignored: %s', einfo)

    def _handleDELMODULE(self, einfo):
        logger.warning('DEPRECATED (DELMODULE) ignored: %s', einfo)

    def _handleADDFMODULE(self, einfo):
        logger.warning('DEPRECATED (ADDFMODULE) ignored: %s', einfo)

    def _handleDELFMODULE(self, einfo):
        logger.warning('DEPRECATED (DELFMODULE) ignored: %s', einfo)

    def _handleADDFUNCTION(self, einfo):
        va, meta = einfo
        self._initFunction(va)

        # node = self._call_graph.addNode( nid=va, repr=self.getName( va ) ) #, color='#00ff00' )
        # node = self._call_graph.getFunctionNode(va, repr=self.getName( va ) )
        node = self._call_graph.getFunctionNode(va)
        self._call_graph.setNodeProp(node, 'repr', self.getName(va))

        # Tell the codeflow subsystem about this one!
        calls_from = meta.get('CallsFrom')
        self.cfctx.addFunctionDef(va, calls_from)

        self.funcmeta[va] = meta

        for name, value in meta.items():
            mcbname = "_fmcb_%s" % name.split(':')[0]
            mcb = getattr(self, mcbname, None)
            if mcb is not None:
                mcb(va, name, value)

    def _handleDELFUNCTION(self, einfo):
        # clear funcmeta, func_args, codeblocks_by_funcva, update codeblocks, blockgraph, locations, etc...
        fva = einfo

        # not every codeblock identifying as this function is stored in funcmeta
        for cb in self.getCodeBlocks():
            if cb[CB_FUNCVA] == fva:
                self._handleDELCODEBLOCK(cb)

        self.funcmeta.pop(fva)
        self.func_args.pop(fva, None)
        self.codeblocks_by_funcva.pop(fva)
        node = self._call_graph.getNode(fva)
        self._call_graph.delNode(node)
        self.cfctx.flushFunction(fva)

        # FIXME: do we want to now seek the function we *should* be in?
        # if xrefs_to, look for non-PROC code xrefs and take their function
        # if the previous instruction falls through, take its function
        # run codeblock analysis on that function to reassociate the blocks
        # with that function

    def _handleSETFUNCMETA(self, einfo):
        funcva, name, value = einfo
        m = self.funcmeta.get(funcva)
        if m is not None:
            m[name] = value
        mcbname = "_fmcb_%s" % name.split(':')[0]
        mcb = getattr(self, mcbname, None)
        if mcb is not None:
            mcb(funcva, name, value)

    def _handleADDCODEBLOCK(self, einfo):
        va,size,funcva = einfo
        self.blockmap.setMapLookup(va, size, einfo)
        self.codeblocks_by_funcva.get(funcva).append(einfo)
        self.codeblocks.append(einfo)

    def _handleDELCODEBLOCK(self, cb):
        va,size,funcva = cb
        self.codeblocks.remove(cb)
        self.codeblocks_by_funcva.get(cb[CB_FUNCVA]).remove(cb)
        self.blockmap.setMapLookup(va, size, None)

    def _handleADDXREF(self, einfo):
        fromva, tova, reftype, rflags = einfo
        xr_to = self.xrefs_by_to.get(tova, None)
        xr_from = self.xrefs_by_from.get(fromva, None)
        if xr_to is None:
            xr_to = []
            self.xrefs_by_to[tova] = xr_to

        if xr_from is None:
            xr_from = []
            self.xrefs_by_from[fromva] = xr_from

        if einfo not in xr_to:  # Just check one for now
            xr_to.append(einfo)
            xr_from.append(einfo)
            self.xrefs.append(einfo)

    def _handleDELXREF(self, einfo):
        fromva, tova, reftype, refflags = einfo
        self.xrefs_by_to[tova].remove(einfo)
        self.xrefs_by_from[fromva].remove(einfo)

    def _handleSETNAME(self, einfo):
        va, name = einfo
        if name is None:
            oldname = self.name_by_va.pop(va, None)
            self.va_by_name.pop(oldname, None)

        else:
            curname = self.name_by_va.get(va)
            if curname is not None:
                logger.debug('replacing 0x%x: %r -> %r', va, curname, name)
                self.va_by_name.pop(curname)

            self.va_by_name[name] = va
            self.name_by_va[va] = name

        if self.isFunction(va):
            fnode = self._call_graph.getFunctionNode(va)
            if name is None:
                self._call_graph.delNodeProp(fnode, 'repr')
            else:
                self._call_graph.setNodeProp(fnode, 'repr', name)

    def _handleADDMMAP(self, einfo):
        if len(einfo) == 5:
            # new "alignment-friendly" event
            va, perms, fname, mbytes, align = einfo
        else:
            # DEPRECATED (21-09-13) - old event style, to support older .viv's
            va, perms, fname, mbytes = einfo
            align = None

        blen = e_mem.MemoryObject.addMemoryMap(self, va, perms, fname, mbytes, align)

        self.locmap.initMapLookup(va, blen)
        self.blockmap.initMapLookup(va, blen)

        # On loading a new memory map, we need to crush a few
        # transmeta items...
        self.transmeta.pop('findPointers',None)

    def _handleDELMMAP(self, mapva):
        e_mem.MemoryObject.delMemoryMap(self, mapva)
        self.locmap.delMapLookup(mapva)
        self.blockmap.delMapLookup(mapva)

        # wipe the opcode cache in case we have cached deleted locations
        self.clearOpcache()

    def _handleADDEXPORT(self, einfo):
        va, etype, name, filename = einfo
        self.exports.append(einfo)
        self.exports_by_va[va] = einfo

    def _handleSETMETA(self, einfo):
        name,value = einfo
        # See if there's a callback handler for this meta set.
        # For "meta namespaces" use the first part to find the
        # callback name....
        mcbname = "_mcb_%s" % name.split(':')[0]
        mcb = getattr(self, mcbname, None)
        if mcb is not None:
            mcb(name, value)
        self.metadata[name] = value

    def _handleCOMMENT(self, einfo):
        va,comment = einfo
        if comment is None:
            self.comments.pop(va, None)
        else:
            self.comments[va] = comment

    def _handleENDIAN(self, einfo):
        self.bigend = einfo
        for arch in self.imem_archs:
            if not arch:
                continue
            arch.setEndian(self.bigend)

    def _handleADDFILE(self, einfo):
        normname, imagebase, md5sum = einfo
        self.filemeta[normname] = {"md5sum":md5sum,"imagebase":imagebase}

    def _handleSETFILEMETA(self, einfo):
        fname, key, value = einfo
        self.filemeta.get(fname)[key] = value

    def _handleADDCOLOR(self, coltup):
        mapname, colmap = coltup
        self.colormaps[mapname] = colmap

    def _handleDELCOLOR(self, mapname):
        self.colormaps.pop(mapname)

    def _handleADDVASET(self, argtup):
        name, defs, rows = argtup
        # NOTE: legacy translation for vaset column types...
        defs = [ (cname,vaset_xlate.get(ctype,ctype)) for (cname,ctype) in defs ]
        self.vasetdefs[name] = defs
        vals = {}
        for row in rows:
            vals[row[0]] = row
        self.vasets[name] = vals

    def _handleDELVASET(self, setname):
        self.vasetdefs.pop(setname)
        self.vasets.pop(setname)

    def _handleADDFREF(self, frtup):
        va, idx, val = frtup
        self.frefs[(va,idx)] = val

    def _handleDELFREF(self, frtup):
        va, idx, val = frtup
        self.frefs.pop((va,idx), None)

    def _handleSETVASETROW(self, argtup):
        name, row = argtup
        self.vasets[name][row[0]] = row

    def _handleDELVASETROW(self, argtup):
        name, va = argtup
        self.vasets[name].pop(va, None)

    def _handleADDFSIG(self, einfo):
        raise NotImplementedError("FSIG is deprecated and should not be used")

    def _handleFOLLOWME(self, va):
        pass

    def _handleCHAT(self, msgtup):
        # FIXME make a GUI window for this...
        user, msg = msgtup
        self.vprint('%s: %s' % (user, msg))

    def _handleSYMHINT(self, msgtup):
        va, idx, hint = msgtup
        if hint is None:
            self.symhints.pop((va,idx), None)
        else:
            self.symhints[(va,idx)] = hint

    def _handleSETFUNCARGS(self, einfo):
        fva, args = einfo
        self.func_args[fva] = args

    def _handleAUTOANALFIN(self, einfo):
        '''
        This event is more for the storage subsystem than anything else.  It
        marks the end of autoanalysis.  Any event beyond this is due to the
        end user or analysis modules they've executed.
        '''
        pass

    def _handleWRITEMEM(self, einfo):
        '''
        Handle permanent writes to a memory map after initialization
        (fname, off, bytez, supv) where supv is supervisor mode...
        '''
        va, bytez, oldbytes = einfo
        with self.getAdminRights():
            e_mem.MemoryObject.writeMemory(self, va, bytez)

    def _initEventHandlers(self):
        self.ehand = [None for x in range(VWE_MAX)]
        self.ehand[VWE_ADDLOCATION] = self._handleADDLOCATION
        self.ehand[VWE_DELLOCATION] = self._handleDELLOCATION
        self.ehand[VWE_ADDSEGMENT] = self._handleADDSEGMENT
        self.ehand[VWE_DELSEGMENT] = None
        self.ehand[VWE_ADDRELOC] = self._handleADDRELOC
        self.ehand[VWE_DELRELOC] = self._handleDELRELOC
        self.ehand[VWE_ADDMODULE] = self._handleADDMODULE
        self.ehand[VWE_DELMODULE] = self._handleDELMODULE
        self.ehand[VWE_ADDFMODULE] = self._handleADDFMODULE
        self.ehand[VWE_DELFMODULE] = self._handleDELFMODULE
        self.ehand[VWE_ADDFUNCTION] = self._handleADDFUNCTION
        self.ehand[VWE_DELFUNCTION] = self._handleDELFUNCTION
        self.ehand[VWE_SETFUNCARGS] = self._handleSETFUNCARGS
        self.ehand[VWE_SETFUNCMETA] = self._handleSETFUNCMETA
        self.ehand[VWE_ADDCODEBLOCK] = self._handleADDCODEBLOCK
        self.ehand[VWE_DELCODEBLOCK] = self._handleDELCODEBLOCK
        self.ehand[VWE_ADDXREF] = self._handleADDXREF
        self.ehand[VWE_DELXREF] = self._handleDELXREF
        self.ehand[VWE_SETNAME] = self._handleSETNAME
        self.ehand[VWE_ADDMMAP] = self._handleADDMMAP
        self.ehand[VWE_DELMMAP] = self._handleDELMMAP
        self.ehand[VWE_ADDEXPORT] = self._handleADDEXPORT
        self.ehand[VWE_DELEXPORT] = None
        self.ehand[VWE_SETMETA] = self._handleSETMETA
        self.ehand[VWE_COMMENT] = self._handleCOMMENT
        self.ehand[VWE_ADDFILE] = self._handleADDFILE
        self.ehand[VWE_DELFILE] = None
        self.ehand[VWE_SETFILEMETA] = self._handleSETFILEMETA
        self.ehand[VWE_ADDCOLOR] = self._handleADDCOLOR
        self.ehand[VWE_DELCOLOR] = self._handleDELCOLOR
        self.ehand[VWE_ADDVASET] = self._handleADDVASET
        self.ehand[VWE_DELVASET] = self._handleDELVASET
        self.ehand[VWE_SETVASETROW] = self._handleSETVASETROW
        self.ehand[VWE_DELVASETROW] = self._handleDELVASETROW
        self.ehand[VWE_ADDFSIG] = self._handleADDFSIG
        self.ehand[VWE_ADDFREF] = self._handleADDFREF
        self.ehand[VWE_DELFREF] = self._handleDELFREF
        self.ehand[VWE_FOLLOWME] = self._handleFOLLOWME
        self.ehand[VWE_CHAT]     = self._handleCHAT
        self.ehand[VWE_SYMHINT]  = self._handleSYMHINT
        self.ehand[VWE_AUTOANALFIN] = self._handleAUTOANALFIN
        self.ehand[VWE_WRITEMEM] = self._handleWRITEMEM
        self.ehand[VWE_ENDIAN] = self._handleENDIAN

        self.thand = [None for x in range(VTE_MAX)]
        self.thand[VTE_IAMLEADER] = self._handleIAMLEADER
        self.thand[VTE_FOLLOWME] = self._handleFOLLOWME
        self.thand[VTE_KILLLEADER] = self._handleKILLLEADER
        self.thand[VTE_MODLEADER] = self._handleMODLEADER

    def _handleFOLLOWME(self, event, einfo):
        uuid, expr = einfo
        logger.debug("_handleFOLLOWME(%r, %r)", event, einfo)
        self.leaderloc[uuid] = expr

    def _handleKILLLEADER(self, event, einfo):
        logger.debug("_handleKILLLEADER(%r, %r)", event, einfo)
        uuid = einfo
        user, fname = self.leaders.pop(uuid)
        self.vprint("*Ended: %s's session '%s' (%r)" % (user,fname,uuid))

    def _handleMODLEADER(self, event, einfo):
        uuid, user, fname = einfo
        self.vprint('*%s changed leader session name to "%s" (%r)' % (user,fname,uuid))

        self.leaders[uuid] = (user, fname)

    def _handleIAMLEADER(self, event, einfo):
        uuid, user, fname, locexpr = einfo
        logger.debug("_handleIAMLEADER(%r, (%r, %r, %r, %r))", event, user, uuid, fname, locexpr)

        self.vprint('*%s invites everyone to follow "%s" (%r)' % (user,fname,uuid))
        self.leaders[uuid] = (user, fname)
        self.leaderloc[uuid] = locexpr

    def _fireEvent(self, event, einfo, local=False, skip=None):
        '''
        Fire an event down the hole.  "local" specifies that this is
        being called on a client (self.server is not None) but we got it
        from the server in the first place so no need to send it back.

        skip is used to tell the server to bypass our channelid when
        putting the event into channel queues (we took care of our own).
        '''

        try:
            if event & VTE_MASK:
                return self._fireTransEvent(event, einfo)

            # Do our main event processing
            self.ehand[event](einfo)

            # If we're supposed to call a server, do that.
            if self.server is not None and local == False:
                self.server._fireEvent(event, einfo, skip=self.rchan)

            # FIXME perhaps we should only process events *via* our server
            # if we have one? Just to confirm it works before we apply it...
            self._event_list.append((event, einfo))

            for id, q in self.chan_lookup.items():
                if id == skip:
                    continue
                try:
                    q.put_nowait((event, einfo))
                except queue.Full as e:
                    logger.warning('Queue is full!')

        except Exception as e:
            logger.error(traceback.format_exc())

    def _fireTransEvent(self, event, einfo):
        for q in self.chan_lookup.values():
            q.put((event, einfo))
        return self.thand[event ^ VTE_MASK](event,einfo)

    def _initFunction(self, funcva):
        # Internal function to initialize all datastructures necessary for
        # a function, but only if they haven't been done already.
        if self.funcmeta.get(funcva) is None:
            self.funcmeta[funcva] = {} # His metadata
            self.codeblocks_by_funcva[funcva] = [] # Init code block list

    #def _loadImportApi(self, apidict):
        #self._imp_api.update( apidict )

    def getEndian(self):
        return self.bigend

    def notifyLoadEvent(self):
        '''
        There are a couple components required before we the "load event" is
        complete.  Each component calls this function, and based on the state
        of other components, the _load_event is set
        '''
        if self.getMeta('Architecture') is None:
            return

        if self.getMeta('Platform') is None:
            return

        self._load_event.set()

#################################################################
#
#  setMeta key callbacks
#
    def _mcb_Architecture(self, name, value):
        archid = envi.getArchByName(value)
        try:
            # Some of the ENVI archs defined may have architecture modules that 
            # are still in progress
            self.setMemArchitecture(archid)
        except IndexError:
            raise ArchModDefException(value) from None

        # This is for legacy stuff...
        #self.arch = envi.getArchModule(value)
        self.arch = self.getMemArchModule()
        self.psize = self.arch.getPointerSize()

        # Default calling convention for architecture
        # This will be superceded by Platform and Parser settings
        defcall = self.arch.getArchDefaultCall()
        if defcall:
            self.setMeta('DefaultCall', defcall)

        self.notifyLoadEvent()

    def _mcb_bigend(self, name, value):
        self.setEndian(bool(value))

    def _mcb_Platform(self, name, value):
        # Default calling convention for platform
        # This supercedes Architecture's setting and should make
        # parser settings obsolete
        defcall = self.arch.getPlatDefaultCall(value)
        if defcall:
            self.setMeta('DefaultCall', defcall)

        self.notifyLoadEvent()

    def _mcb_FileBytes(self, name, value):
        if not self.parsedbin:
            byts = viv_parsers.uncompressBytes(value)
            fmt = viv_parsers.guessFormat(byts)
            parser = viv_parsers.getBytesParser(fmt)
            if parser:
                self.parsedbin = parser(byts)

    def _mcb_ustruct(self, name, ssrc):
        # All meta values in the "ustruct" namespace are user defined
        # structure defintions in C.
        sname = name.split(':')[1]
        ctor = vs_cparse.ctorFromCSource( ssrc )
        self.vsbuilder.addVStructCtor( sname, ctor )

    def _mcb_GUID(self, name, guid):
        self._load_guid.set()

    def _mcb_WorkspaceServer(self, name, wshost):
        self.vprint('Workspace was Saved to Server: %s' % wshost)
        self.vprint('(You must close this local copy and work from the server to stay in sync.)')

    def _fmcb_Thunk(self, funcva, th, thunkname):
        # If the function being made a thunk is registered
        # in NoReturnApis, update codeflow...
        if self.getMeta('NoReturnApis').get(thunkname.lower()):
            self.cfctx.addNoReturnAddr(funcva)

    def _fmcb_CallsFrom(self, funcva, th, callsfrom):
        for va in callsfrom:
            f2va = self.getFunction( va )
            if f2va is not None:
                self._call_graph.getCallEdge( funcva, f2va )

    def _fmcb_LocalSymbol(self, fva, mname, locsym):
        fva,spdelta,symtype,syminfo = locsym
        self.localsyms[fva][spdelta] = locsym

def trackDynBranches(cfctx, op, vw, bflags, branches):
    '''
    track dynamic branches
    '''
    # FIXME: do we want to filter anything out?  
    #  jmp edx
    #  jmp dword [ebx + 68]
    #  call eax
    #  call dword [ebx + eax * 4 - 228]

    # if we have any xrefs from here, we have already been analyzed.  nevermind.
    if len(vw.getXrefsFrom(op.va)):
        return

    logger.info("0x%x: Dynamic Branch found:  %s" % (op.va, op))
    vw.setVaSetRow('DynamicBranches', (op.va, repr(op), bflags))

class VivCodeFlowContext(e_codeflow.CodeFlowContext):
    def __init__(self, mem, persist=False, exptable=True, recurse=True):
        e_codeflow.CodeFlowContext.__init__(self, mem, persist=persist, exptable=exptable, recurse=recurse)
        self.addDynamicBranchHandler(trackDynBranches)

    def _cb_noflow(self, srcva, dstva):
        vw = self._mem
        loc = vw.getLocation( srcva )
        if loc is None:
            return

        lva,lsize,ltype,linfo = loc
        if ltype != LOC_OP:
            return

        # Update the location def for NOFALL bit
        vw.delLocation(lva)
        vw.addLocation(lva, lsize, ltype, linfo | envi.IF_NOFALL)

        vw.setVaSetRow('NoReturnCalls', (lva,))

    # NOTE: self._mem is the viv workspace...
    def _cb_opcode(self, va, op, branches):
        '''
        callback for each OPCODE in codeflow analysis
        must return list of branches, modified for our purposes
        '''
        loc = self._mem.getLocation(va)
        if loc is None:

            # dont code flow through import calls
            branches = [br for br in branches if not self._mem.isLocType(br[0], LOC_IMPORT)]

            self._mem.makeOpcode(op.va, op=op)
            # TODO: future home of makeOpcode branch/xref analysis
            if not self._mem.isNoReturnVa(op.va):
                return branches

        elif loc[L_LTYPE] != LOC_OP:
            locrepr = self._mem.reprLocation(loc)
            logger.warning("_cb_opcode(0x%x): LOCATION ALREADY EXISTS: loc: %r", va, locrepr)

        return ()

    def _cb_function(self, fva, fmeta):

        vw = self._mem
        if vw.isFunction(fva):
            return

        # This may be possible if an export/symbol was mistaken for
        # a function...
        if not vw.isLocType(fva, LOC_OP):
            return

        # If the function doesn't have a name, make one
        if vw.getName(fva) is None:
            vw.makeName(fva, "sub_%.8x" % fva)

        vw._fireEvent(VWE_ADDFUNCTION, (fva,fmeta))

        # Go through the function analysis modules in order
        vw.analyzeFunction(fva)

        fname = vw.getName( fva )
        if vw.getMeta('NoReturnApis').get( fname.lower() ):
            self._cf_noret[fva] = True

        if len( vw.getFunctionBlocks( fva )) == 1:
            return

        fmeta = vw.getFunctionMetaDict(fva)
        for lva in vw.getVaSetRows('NoReturnCalls'):
            va = lva[0]
            ctup = vw.getCodeBlock(va)
            if ctup and fva == ctup[2] and vw.getFunctionMeta(fva, 'BlockCount', default=0) == 1:
                self._cf_noret[fva] = True
                break

    def _cb_branchtable(self, tablebase, tableva, destva):

        if tablebase != tableva and self._mem.getXrefsTo(tableva):
            return False

        if self._mem.getLocation(tableva) is None:
            self._mem.makePointer(tableva, tova=destva, follow=False)

        return True




'''
Some utilities for debugging on android.
'''
import os
import subprocess

import vdb.release as vdb_release


utildir = os.path.dirname(__file__)
droidtmp = '/data/local/tmp/'


def adbCommand(*argv):
    return subprocess.check_output(argv, stderr=subprocess.PIPE)


message = '''
Due to silly/complicated reasons, you have to issue the
vdb command from within adb yourself (ask visi to rage about
it if youd like to know why...)

Please run this:
cd %s; sh androidpy.sh vdb.pyz
''' % droidtmp


def runVdbOnDroid():
    zfile = os.path.join(utildir, 'vdb.pyz')
    script = os.path.join(utildir, 'androidpy.sh')
    vdb_release.getSourceZip(fname=zfile)

    # copy in the script and the source zip
    print('Pushing vdb.pyz...')
    adbCommand('adb', 'push', zfile, droidtmp + 'vdb.pyz')
    print('Pushing python shell script...')
    adbCommand('adb', 'push', script, droidtmp + 'androidpy.sh')
    print(message)
    # os.system('adb shell sh %s/androidpy.sh %s/vdb.pyz' % ( droidtmp, droidtmp ))
    os.system('adb shell')


if __name__ == '__main__':
    runVdbOnDroid()



"""
Unified expression helpers.
"""


class ExpressionFail(Exception):
    def __init__(self, pycode, exception):
        Exception.__init__(self)
        self.pycode = pycode
        self.exception = exception

    def __repr__(self):
        return "ExpressionFail: %r is not a valid expression in this context (%r)" % \
                (self.pycode, self.exception)

    def __str__(self):
        return self.__repr__()


def evaluate(pycode, locvars):
    try:
        val = eval(pycode, {}, locvars)
    except Exception:
        try:
            # check through the keys for anything we might want to replace
            keys = list(locvars.keys())

            # sort the keys in reverse order so that longer matching strings take priority
            keys.sort(reverse=True)

            # replace the substrings with the string versions of the lookup value
            for key in keys:
                if key in pycode:
                    pval = locvars[key]
                    pycode = pycode.replace(key, str(pval))

            val = eval(pycode, {}, locvars)

        except Exception as e:
            raise ExpressionFail(pycode, e)

    return val


class ExpressionLocals(dict):
    """
    An object to act as the locals dictionary for the evaluation
    of envi expressions.  You may pass in an envi.symstore.resolver.SymbolResolver
    object to automagically use symbols in your expressions.
    """
    def __init__(self, symobj=None):
        dict.__init__(self)
        self.symobj = symobj

    def __getitem__(self, name):
        if self.symobj is not None:
            ret = self.symobj.getSymByName(name)
            if ret is not None:
                if ret.symtype == 3:
                    return ret
                else:
                    return ret.value
        return dict.__getitem__(self, name)

    get = __getitem__

    def __iter__(self):
        if self.symobj is not None:
            for va, name in self.symobj.getNames():
                yield name

        yield from dict.__iter__(self)

    def keys(self):
        return [key for key in self]

    def __contains__(self, key):
        return self.__getitem__(key) is not None

class MemoryExpressionLocals(ExpressionLocals):

    def __init__(self, memobj, symobj=None):
        ExpressionLocals.__init__(self, symobj=symobj)
        self.memobj = memobj
        self.update({
            'mapbase': self.mapbase,
            'maplen': self.maplen,
            'ispoi': self.ispoi,
            'mem': self.mem,
            'poi': self.poi,
            'sym': self.sym,
        })

    def sym(self, symstr):
        '''
        An easy to use utility for symbols which have un-pythonic names.

        Example x = sym('kernel32.??2@$$FYAPAXI@Z')
        '''
        return int(evaluate(symstr, self))

    def mapbase(self, address):
        """
        The expression mapbase(address) returns the base address of the
        memory mapped area containing "address"
        """
        map = self.memobj.getMemoryMap(address)
        if not map:
            raise Exception("ERROR - un-mapped address in mapbase()")
        return map[0]

    def maplen(self, address):
        """
        The expression maplen(address) returns the length of the
        memory mapped area containing "address".
        """
        map = self.memobj.getMemoryMap(address)
        if not map:
            raise Exception("ERROR - un-mapped address in maplen()")
        return map[1]

    def ispoi(self, addr):
        """
        The expression ispoi(value) returns True if the specified value
        is a valid pointer.  Otherwise, False.
        """
        return self.memobj.isValidPointer(addr)

    def mem(self, addr, size):
        """
        Read and return memory.

        Example: mem(ecx, 20)
        """
        return self.memobj.readMemory(addr, size)

    def poi(self, address):
        """
        When expression contains "poi(addr)" this will return
        the address pointed to by addr.
        """
        return self.memobj.readMemoryPtr(address)



'''
Specialized breakpoints which identify dangerous calling
mechanisms and tag them.
'''
import logging

import envi.memory as e_mem
import vtrace.breakpoints as vt_breakpoints

logger = logging.getLogger(__name__)

def getStackArg(trace, argidx):
    '''
    Assuming we are at the instruction after
    a call, grab the argument at the specified
    index (skipping the saved instruction pointer).
    '''
    cc = detect_cc(trace)
    args = cc.getCallArgs(trace, argidx)
    return args[-1]

class SniperDynArgBreak(vt_breakpoints.Breakpoint):
    '''
    A breakpoint for use in determining if an API was called
    with a dynamic pointer.
    '''

    def __init__(self, symname, argidx):
        vt_breakpoints.Breakpoint.__init__(self, None, expression=symname)
        self.fastbreak = True
        self._argidx = argidx
        self._symname = symname

    def getName(self):
        return "%s argidx: %d" % (self._symname, self._argidx)

    def notify(self, event, trace):
        arg = getArg(trace, self._argidx)
        self.fastbreak = True
        if trace.probeMemory(arg, 1, e_mem.MM_WRITE):
            logger.info('SNIPER: %s TOOK DYNAMIC ARG IDX %d (0x%.8x)', self._symname, self._argidx, arg)
            self.fastbreak = False


class SniperArgValueBreak(vt_breakpoints.Breakpoint):
    '''
    A breakpoint for monitoring an API for being called with a particular
    value.
    '''
    def __init__(self, symname, argidx, argval):
        vt_breakpoints.Breakpoint.__init__(self, None, expression=symname)
        self.fastbreak = True
        self._argval = argval
        self._argidx = argidx
        self._symname = symname

    def notify(self, event, trace):
        arg = getArg(trace, self._argidx)
        self.fastbreak = True
        if(arg == self._argval):
            print("SNIPER: %s TOOK VALUE (0x%.8x) FOR ARG IDX %d" % (self._symname, arg, self._argidx))
            self.fastbreak = False


def snipeDynArg(trace, symname, argidx):
    '''
    Construct a SnyperDynArgBreak and snap it in.
    '''
    bp = SniperDynArgBreak(symname, argidx)
    bpid = trace.addBreakpoint(bp)
    return bpid

def snipeArgValue(trace, symname, argidx, argval):
    '''
    Construct a SnyperArgValueBreak and snap it in.
    '''
    bp = SniperArgValueBreak(symname, argidx, argval)
    bpid = trace.addBreakpoint(bp)
    return bpid



# APIs for Windows 32-bit MSVC 2012 runtime library (msvcr110).
# Format:  rettype, retname, callconv, exactname, arglist(type, name)
#          arglist type is one of ['int', 'void *']
#          arglist name is one of [None, 'funcptr', 'obj', 'ptr']
#
# The subset of functions defined below accounts for 80% of observed calls.

api_defs = {
    'msvcr110.??2@yapaxi@z':( 'int', None, 'cdecl', 'msvcr110.??2@YAPAXI@Z', (('int', None),) ),
    'msvcr110.??_u@yapaxi@z':( 'int', None, 'cdecl', 'msvcr110.??_U@YAPAXI@Z', (('int', None),) ),
    'msvcr110.??3@yaxpax@z':( 'void', None, 'cdecl', 'msvcr110.??3@YAXPAX@Z', (('void *', 'ptr'),) ),
    'msvcr110.??_v@yaxpax@z':( 'void', None, 'cdecl', 'msvcr110.??_V@YAXPAX@Z', (('void *', 'ptr'),) ),

    'msvcr110._cxxthrowexception':( 'void', None, 'stdcall', 'msvcr110._CxxThrowException', (('void *', 'ptr'), ('int', None)) ),
    'msvcr110.__iob_func':( 'int', None, 'cdecl', 'msvcr110.__iob_func', () ),
    'msvcr110._invalid_parameter_noinfo':( 'void', None, 'cdecl', 'msvcr110._invalid_parameter_noinfo', () ),
    'msvcr110._errno':( 'int', None, 'cdecl', 'msvcr110._errno', () ),

    'msvcr110._exit':( 'int', None, 'cdecl', 'msvcr110._exit', (('int', None),) ),
    'msvcr110.abort':( 'int', None, 'cdecl', 'msvcr110.abort', () ),
    'msvcr110.exit':( 'int', None, 'cdecl', 'msvcr110.exit', (('int', None),) ),
    'msvcr110.quick_exit':( 'int', None, 'cdecl', 'msvcr110.quick_exit', (('int', None),) ),

    'msvcr110.calloc':( 'int', None, 'cdecl', 'msvcr110.calloc', (('int', None), ('int', None)) ),
    'msvcr110.malloc':( 'int', None, 'cdecl', 'msvcr110.malloc', (('int', None),) ),
    'msvcr110.free':( 'void', None, 'cdecl', 'msvcr110.free', (('void *', 'ptr'),) ),

    'msvcr110.memcmp':( 'int', None, 'cdecl', 'msvcr110.memcmp', (('void *', 'ptr'), ('void *', 'ptr'), ('int', None)) ),
    'msvcr110.memcpy':( 'int', None, 'cdecl', 'msvcr110.memcpy', (('void *', 'ptr'), ('void *', 'ptr'), ('int', None)) ),
    'msvcr110.memmove':( 'int', None, 'cdecl', 'msvcr110.memmove', (('void *', 'ptr'), ('void *', 'ptr'), ('int', None)) ),
    'msvcr110.memmove_s':( 'int', None, 'cdecl', 'msvcr110.memmove_s', (('void *', 'ptr'), ('int', None), ('void *', 'ptr'), ('int', None)) ),
    'msvcr110.memset':( 'int', None, 'cdecl', 'msvcr110.memset', (('void *', 'ptr'), ('int', None), ('int', None)) ),

    'msvcr110._snprintf':( 'int', None, 'cdecl', 'msvcr110._snprintf', (('int', None), ('int', None), ('int', None)) ),
    'msvcr110._swprintf':( 'int', None, 'cdecl', 'msvcr110._snprintf', (('int', None), ('int', None), ('int', None), ('int', None)) ),
    'msvcr110.fprintf':( 'int', None, 'cdecl', 'msvcr110.fprintf', (('void *', 'ptr'), ('int', None)) ),
    'msvcr110.sprintf':( 'int', None, 'cdecl', 'msvcr110.sprintf', (('int', None), ('int', None)) ),
    'msvcr110.printf':( 'int', None, 'cdecl', 'msvcr110.printf', (('int', None),) ),

    'msvcr110._stricmp':( 'int', None, 'cdecl', 'msvcr110._stricmp', (('void *', 'ptr'), ('void *', 'ptr')) ),
    'msvcr110.strchr':( 'int', None, 'cdecl', 'msvcr110.strchr', (('void *', 'ptr'), ('int', None)) ),
    'msvcr110.strcmp':( 'int', None, 'cdecl', 'msvcr110.strcmp', (('void *', 'ptr'), ('void *', 'ptr')) ),
    'msvcr110.strcpy':( 'int', None, 'cdecl', 'msvcr110.strcpy', (('void *', 'ptr'), ('void *', 'ptr')) ),
    'msvcr110.strlen':( 'int', None, 'cdecl', 'msvcr110.strlen', (('void *', 'ptr'),) ),
    'msvcr110.strncmp':( 'int', None, 'cdecl', 'msvcr110.strncmp', (('void *', 'ptr'), ('void *', 'ptr'), ('int', None)) ),
    'msvcr110.strncpy':( 'int', None, 'cdecl', 'msvcr110.strncpy', (('void *', 'ptr'), ('void *', 'ptr'), ('int', None)) ),
    'msvcr110.strstr':( 'int', None, 'cdecl', 'msvcr110.strstr', (('void *', 'ptr'), ('void *', 'ptr')) ),
    'msvcr110._wcsicmp':( 'int', None, 'cdecl', 'msvcr110._wcsicmp', (('void *', 'ptr'), ('void *', 'ptr')) ),
    'msvcr110.wcslen':( 'int', None, 'cdecl', 'msvcr110.wcslen', (('void *', 'ptr'),) ),
    'msvcr110.wcsrchr':( 'int', None, 'cdecl', 'msvcr110.wcsrchr', (('void *', 'ptr'), ('int', None)) ),

    'msvcr110._time64':( 'int', None, 'cdecl', 'msvcr110._time64', (('void *', 'ptr'),) ),
    'msvcr110.atoi':( 'int', None, 'cdecl', 'msvcr110.atoi', (('void *', 'ptr'),) ),
    'msvcr110.getenv':( 'int', None, 'cdecl', 'msvcr110.getenv', (('void *', 'ptr'),) ),
    'msvcr110.fclose':( 'int', None, 'cdecl', 'msvcr110.fclose', (('void *', 'ptr'),) ),
    'msvcr110.bsearch':( 'int', None, 'cdecl', 'msvcr110.bsearch', (('void *', 'ptr'), ('void *', 'ptr'), ('int', None), ('int', None), ('void *', 'funcptr')) )
    }



from PyQt5.QtWidgets import QWidget

import vqt.saveable as vq_save

import vtrace.qt

class VdbWidgetWindow(QWidget, vq_save.SaveableWidget, vtrace.qt.VQTraceNotifier):
    '''
    a base window class for widgets to inherit from for vdb.
    this gives your window/widget access to the vdb instance (self.db), the gui
    instance (self.db.gui), and the persistent trace object (self.dbt).

    implement vqLoad for tracer events.
    implement vdbUIEvent for events caused by user interaction.
    state between runs of the debugger.
    '''
    def __init__(self, db, dbt, parent=None):
        QWidget.__init__(self, parent=parent)
        vq_save.SaveableWidget.__init__(self)
        vtrace.qt.VQTraceNotifier.__init__(self, trace=dbt)

        self.db = db
        self.dbt = dbt

    def keyPressEvent(self, event):
        '''
        handle the global hotkeys.
        '''
        self.db.gui.keyPressEvent(event)



ord_names = {
    2:'SysAllocString',
    3:'SysReAllocString',
    4:'SysAllocStringLen',
    5:'SysReAllocStringLen',
    6:'SysFreeString',
    7:'SysStringLen',
    8:'VariantInit',
    9:'VariantClear',
    10:'VariantCopy',
    11:'VariantCopyInd',
    12:'VariantChangeType',
    13:'VariantTimeToDosDateTime',
    14:'DosDateTimeToVariantTime',
    15:'SafeArrayCreate',
    16:'SafeArrayDestroy',
    17:'SafeArrayGetDim',
    18:'SafeArrayGetElemsize',
    19:'SafeArrayGetUBound',
    20:'SafeArrayGetLBound',
    21:'SafeArrayLock',
    22:'SafeArrayUnlock',
    23:'SafeArrayAccessData',
    24:'SafeArrayUnaccessData',
    25:'SafeArrayGetElement',
    26:'SafeArrayPutElement',
    27:'SafeArrayCopy',
    28:'DispGetParam',
    29:'DispGetIDsOfNames',
    30:'DispInvoke',
    31:'CreateDispTypeInfo',
    32:'CreateStdDispatch',
    33:'RegisterActiveObject',
    34:'RevokeActiveObject',
    35:'GetActiveObject',
    36:'SafeArrayAllocDescriptor',
    37:'SafeArrayAllocData',
    38:'SafeArrayDestroyDescriptor',
    39:'SafeArrayDestroyData',
    40:'SafeArrayRedim',
    41:'SafeArrayAllocDescriptorEx',
    42:'SafeArrayCreateEx',
    43:'SafeArrayCreateVectorEx',
    44:'SafeArraySetRecordInfo',
    45:'SafeArrayGetRecordInfo',
    46:'VarParseNumFromStr',
    47:'VarNumFromParseNum',
    48:'VarI2FromUI1',
    49:'VarI2FromI4',
    50:'VarI2FromR4',
    51:'VarI2FromR8',
    52:'VarI2FromCy',
    53:'VarI2FromDate',
    54:'VarI2FromStr',
    55:'VarI2FromDisp',
    56:'VarI2FromBool',
    57:'SafeArraySetIID',
    58:'VarI4FromUI1',
    59:'VarI4FromI2',
    60:'VarI4FromR4',
    61:'VarI4FromR8',
    62:'VarI4FromCy',
    63:'VarI4FromDate',
    64:'VarI4FromStr',
    65:'VarI4FromDisp',
    66:'VarI4FromBool',
    67:'SafeArrayGetIID',
    68:'VarR4FromUI1',
    69:'VarR4FromI2',
    70:'VarR4FromI4',
    71:'VarR4FromR8',
    72:'VarR4FromCy',
    73:'VarR4FromDate',
    74:'VarR4FromStr',
    75:'VarR4FromDisp',
    76:'VarR4FromBool',
    77:'SafeArrayGetVartype',
    78:'VarR8FromUI1',
    79:'VarR8FromI2',
    80:'VarR8FromI4',
    81:'VarR8FromR4',
    82:'VarR8FromCy',
    83:'VarR8FromDate',
    84:'VarR8FromStr',
    85:'VarR8FromDisp',
    86:'VarR8FromBool',
    87:'VarFormat',
    88:'VarDateFromUI1',
    89:'VarDateFromI2',
    90:'VarDateFromI4',
    91:'VarDateFromR4',
    92:'VarDateFromR8',
    93:'VarDateFromCy',
    94:'VarDateFromStr',
    95:'VarDateFromDisp',
    96:'VarDateFromBool',
    97:'VarFormatDateTime',
    98:'VarCyFromUI1',
    99:'VarCyFromI2',
    100:'VarCyFromI4',
    101:'VarCyFromR4',
    102:'VarCyFromR8',
    103:'VarCyFromDate',
    104:'VarCyFromStr',
    105:'VarCyFromDisp',
    106:'VarCyFromBool',
    107:'VarFormatNumber',
    108:'VarBstrFromUI1',
    109:'VarBstrFromI2',
    110:'VarBstrFromI4',
    111:'VarBstrFromR4',
    112:'VarBstrFromR8',
    113:'VarBstrFromCy',
    114:'VarBstrFromDate',
    115:'VarBstrFromDisp',
    116:'VarBstrFromBool',
    117:'VarFormatPercent',
    118:'VarBoolFromUI1',
    119:'VarBoolFromI2',
    120:'VarBoolFromI4',
    121:'VarBoolFromR4',
    122:'VarBoolFromR8',
    123:'VarBoolFromDate',
    124:'VarBoolFromCy',
    125:'VarBoolFromStr',
    126:'VarBoolFromDisp',
    127:'VarFormatCurrency',
    128:'VarWeekdayName',
    129:'VarMonthName',
    130:'VarUI1FromI2',
    131:'VarUI1FromI4',
    132:'VarUI1FromR4',
    133:'VarUI1FromR8',
    134:'VarUI1FromCy',
    135:'VarUI1FromDate',
    136:'VarUI1FromStr',
    137:'VarUI1FromDisp',
    138:'VarUI1FromBool',
    139:'VarFormatFromTokens',
    140:'VarTokenizeFormatString',
    141:'VarAdd',
    142:'VarAnd',
    143:'VarDiv',
    144:'DllCanUnloadNow',
    145:'DllGetClassObject',
    146:'DispCallFunc',
    147:'VariantChangeTypeEx',
    148:'SafeArrayPtrOfIndex',
    149:'SysStringByteLen',
    150:'SysAllocStringByteLen',
    151:'DllRegisterServer',
    152:'VarEqv',
    153:'VarIdiv',
    154:'VarImp',
    155:'VarMod',
    156:'VarMul',
    157:'VarOr',
    158:'VarPow',
    159:'VarSub',
    160:'CreateTypeLib',
    161:'LoadTypeLib',
    162:'LoadRegTypeLib',
    163:'RegisterTypeLib',
    164:'QueryPathOfRegTypeLib',
    165:'LHashValOfNameSys',
    166:'LHashValOfNameSysA',
    167:'VarXor',
    168:'VarAbs',
    169:'VarFix',
    170:'OaBuildVersion',
    171:'ClearCustData',
    172:'VarInt',
    173:'VarNeg',
    174:'VarNot',
    175:'VarRound',
    176:'VarCmp',
    177:'VarDecAdd',
    178:'VarDecDiv',
    179:'VarDecMul',
    180:'CreateTypeLib2',
    181:'VarDecSub',
    182:'VarDecAbs',
    183:'LoadTypeLibEx',
    184:'SystemTimeToVariantTime',
    185:'VariantTimeToSystemTime',
    186:'UnRegisterTypeLib',
    187:'VarDecFix',
    188:'VarDecInt',
    189:'VarDecNeg',
    190:'VarDecFromUI1',
    191:'VarDecFromI2',
    192:'VarDecFromI4',
    193:'VarDecFromR4',
    194:'VarDecFromR8',
    195:'VarDecFromDate',
    196:'VarDecFromCy',
    197:'VarDecFromStr',
    198:'VarDecFromDisp',
    199:'VarDecFromBool',
    200:'GetErrorInfo',
    201:'SetErrorInfo',
    202:'CreateErrorInfo',
    203:'VarDecRound',
    204:'VarDecCmp',
    205:'VarI2FromI1',
    206:'VarI2FromUI2',
    207:'VarI2FromUI4',
    208:'VarI2FromDec',
    209:'VarI4FromI1',
    210:'VarI4FromUI2',
    211:'VarI4FromUI4',
    212:'VarI4FromDec',
    213:'VarR4FromI1',
    214:'VarR4FromUI2',
    215:'VarR4FromUI4',
    216:'VarR4FromDec',
    217:'VarR8FromI1',
    218:'VarR8FromUI2',
    219:'VarR8FromUI4',
    220:'VarR8FromDec',
    221:'VarDateFromI1',
    222:'VarDateFromUI2',
    223:'VarDateFromUI4',
    224:'VarDateFromDec',
    225:'VarCyFromI1',
    226:'VarCyFromUI2',
    227:'VarCyFromUI4',
    228:'VarCyFromDec',
    229:'VarBstrFromI1',
    230:'VarBstrFromUI2',
    231:'VarBstrFromUI4',
    232:'VarBstrFromDec',
    233:'VarBoolFromI1',
    234:'VarBoolFromUI2',
    235:'VarBoolFromUI4',
    236:'VarBoolFromDec',
    237:'VarUI1FromI1',
    238:'VarUI1FromUI2',
    239:'VarUI1FromUI4',
    240:'VarUI1FromDec',
    241:'VarDecFromI1',
    242:'VarDecFromUI2',
    243:'VarDecFromUI4',
    244:'VarI1FromUI1',
    245:'VarI1FromI2',
    246:'VarI1FromI4',
    247:'VarI1FromR4',
    248:'VarI1FromR8',
    249:'VarI1FromDate',
    250:'VarI1FromCy',
    251:'VarI1FromStr',
    252:'VarI1FromDisp',
    253:'VarI1FromBool',
    254:'VarI1FromUI2',
    255:'VarI1FromUI4',
    256:'VarI1FromDec',
    257:'VarUI2FromUI1',
    258:'VarUI2FromI2',
    259:'VarUI2FromI4',
    260:'VarUI2FromR4',
    261:'VarUI2FromR8',
    262:'VarUI2FromDate',
    263:'VarUI2FromCy',
    264:'VarUI2FromStr',
    265:'VarUI2FromDisp',
    266:'VarUI2FromBool',
    267:'VarUI2FromI1',
    268:'VarUI2FromUI4',
    269:'VarUI2FromDec',
    270:'VarUI4FromUI1',
    271:'VarUI4FromI2',
    272:'VarUI4FromI4',
    273:'VarUI4FromR4',
    274:'VarUI4FromR8',
    275:'VarUI4FromDate',
    276:'VarUI4FromCy',
    277:'VarUI4FromStr',
    278:'VarUI4FromDisp',
    279:'VarUI4FromBool',
    280:'VarUI4FromI1',
    281:'VarUI4FromUI2',
    282:'VarUI4FromDec',
    283:'BSTR_UserSize',
    284:'BSTR_UserMarshal',
    285:'BSTR_UserUnmarshal',
    286:'BSTR_UserFree',
    287:'VARIANT_UserSize',
    288:'VARIANT_UserMarshal',
    289:'VARIANT_UserUnmarshal',
    290:'VARIANT_UserFree',
    291:'LPSAFEARRAY_UserSize',
    292:'LPSAFEARRAY_UserMarshal',
    293:'LPSAFEARRAY_UserUnmarshal',
    294:'LPSAFEARRAY_UserFree',
    295:'LPSAFEARRAY_Size',
    296:'LPSAFEARRAY_Marshal',
    297:'LPSAFEARRAY_Unmarshal',
    298:'VarDecCmpR8',
    299:'VarCyAdd',
    300:'DllUnregisterServer',
    301:'OACreateTypeLib2',
    303:'VarCyMul',
    304:'VarCyMulI4',
    305:'VarCySub',
    306:'VarCyAbs',
    307:'VarCyFix',
    308:'VarCyInt',
    309:'VarCyNeg',
    310:'VarCyRound',
    311:'VarCyCmp',
    312:'VarCyCmpR8',
    313:'VarBstrCat',
    314:'VarBstrCmp',
    315:'VarR8Pow',
    316:'VarR4CmpR8',
    317:'VarR8Round',
    318:'VarCat',
    319:'VarDateFromUdateEx',
    322:'GetRecordInfoFromGuids',
    323:'GetRecordInfoFromTypeInfo',
    325:'SetVarConversionLocaleSetting',
    326:'GetVarConversionLocaleSetting',
    327:'SetOaNoCache',
    329:'VarCyMulI8',
    330:'VarDateFromUdate',
    331:'VarUdateFromDate',
    332:'GetAltMonthNames',
    333:'VarI8FromUI1',
    334:'VarI8FromI2',
    335:'VarI8FromR4',
    336:'VarI8FromR8',
    337:'VarI8FromCy',
    338:'VarI8FromDate',
    339:'VarI8FromStr',
    340:'VarI8FromDisp',
    341:'VarI8FromBool',
    342:'VarI8FromI1',
    343:'VarI8FromUI2',
    344:'VarI8FromUI4',
    345:'VarI8FromDec',
    346:'VarI2FromI8',
    347:'VarI2FromUI8',
    348:'VarI4FromI8',
    349:'VarI4FromUI8',
    360:'VarR4FromI8',
    361:'VarR4FromUI8',
    362:'VarR8FromI8',
    363:'VarR8FromUI8',
    364:'VarDateFromI8',
    365:'VarDateFromUI8',
    366:'VarCyFromI8',
    367:'VarCyFromUI8',
    368:'VarBstrFromI8',
    369:'VarBstrFromUI8',
    370:'VarBoolFromI8',
    371:'VarBoolFromUI8',
    372:'VarUI1FromI8',
    373:'VarUI1FromUI8',
    374:'VarDecFromI8',
    375:'VarDecFromUI8',
    376:'VarI1FromI8',
    377:'VarI1FromUI8',
    378:'VarUI2FromI8',
    379:'VarUI2FromUI8',
    401:'OleLoadPictureEx',
    402:'OleLoadPictureFileEx',
    411:'SafeArrayCreateVector',
    412:'SafeArrayCopyData',
    413:'VectorFromBstr',
    414:'BstrFromVector',
    415:'OleIconToCursor',
    416:'OleCreatePropertyFrameIndirect',
    417:'OleCreatePropertyFrame',
    418:'OleLoadPicture',
    419:'OleCreatePictureIndirect',
    420:'OleCreateFontIndirect',
    421:'OleTranslateColor',
    422:'OleLoadPictureFile',
    423:'OleSavePictureFile',
    424:'OleLoadPicturePath',
    425:'VarUI4FromI8',
    426:'VarUI4FromUI8',
    427:'VarI8FromUI8',
    428:'VarUI8FromI8',
    429:'VarUI8FromUI1',
    430:'VarUI8FromI2',
    431:'VarUI8FromR4',
    432:'VarUI8FromR8',
    433:'VarUI8FromCy',
    434:'VarUI8FromDate',
    435:'VarUI8FromStr',
    436:'VarUI8FromDisp',
    437:'VarUI8FromBool',
    438:'VarUI8FromI1',
    439:'VarUI8FromUI2',
    440:'VarUI8FromUI4',
    441:'VarUI8FromDec',
    442:'RegisterTypeLibForUser',
    443:'UnRegisterTypeLibForUser',
}



import os
import sys

import vstruct.qt as vs_qt
import envi.expression as e_expr
import envi.qt.config as e_q_config

import vqt.cli as vq_cli
import vqt.main as vq_main
import vqt.colors as vq_colors
import vqt.qpython as vq_python
import vqt.hotkeys as vq_hotkey
import vqt.application as vq_app

import vivisect.cli as viv_cli
import vivisect.base as viv_base
import vivisect.vdbext as viv_vdbext
import vivisect.qt.tips as viv_q_tips
import vivisect.qt.views as viv_q_views
import vivisect.qt.memory as viv_q_memory
import vivisect.qt.remote as viv_q_remote
import vivisect.qt.ustruct as viv_q_ustruct
import vivisect.extensions as viv_extensions
import vivisect.qt.funcgraph as viv_q_funcgraph
import vivisect.qt.funcviews as viv_q_funcviews
import vivisect.qt.symboliks as viv_q_symboliks
import vivisect.remote.share as viv_share
import vivisect.analysis.generic.symswitchcase as symswitch

from PyQt5 import QtCore
from PyQt5.QtWidgets import QInputDialog

from vqt.common import *
from vivisect.const import *
from envi.threads import firethread
from vqt.main import getOpenFileName, getSaveFileName
from vqt.saveable import compat_isNone

dock_top = QtCore.Qt.TopDockWidgetArea
dock_right = QtCore.Qt.RightDockWidgetArea


class VQVivMainWindow(viv_base.VivEventDist, vq_app.VQMainCmdWindow):

    # Child windows may emit this on "navigate" requests...
    # vivNavSignal = QtCore.pyqtSignal(str, name='vivNavSignal')
    vivMemColorSignal = QtCore.pyqtSignal(dict, name='vivMemColorSignal')

    def __init__(self, vw):
        self.vw = vw
        vw._viv_gui = self
        # DEV: hijack the workspace's vprint so that they get routed to the UI canvas
        # and not out to the stdout
        vw.vprint = self.vprint
        viv_base.VivEventDist.__init__(self, vw=vw)
        vq_app.VQMainCmdWindow.__init__(self, 'Vivisect', vw)

        self.vqAddMenuField('&File.Open', self._menuFileOpen)
        self.vqAddMenuField('&File.Save', self._menuFileSave)
        self.vqAddMenuField('&File.Save As', self._menuFileSaveAs)
        self.vqAddMenuField('&File.Save to Server', self._menuFileSaveServer)

        self.vqAddMenuField('&File.Quit', self.close)
        self.vqAddMenuField('&Edit.&Preferences', self._menuEditPrefs)

        self.vqAddMenuField('&View.&Exports', self._menuViewExports)
        self.vqAddMenuField('&View.&Functions', self._menuViewFunctions)
        self.vqAddMenuField('&View.&Imports', self._menuViewImports)
        self.vqAddMenuField('&View.&Names', self._menuViewNames)
        self.vqAddMenuField('&View.&Memory', self._menuViewMemory)
        self.vqAddMenuField('&View.&Function Graph', self._menuViewFuncGraph)
        self.vqAddMenuField('&View.&Strings', self._menuViewStrings)
        self.vqAddMenuField('&View.&Structures', self._menuViewStructs)
        self.vqAddMenuField('&View.&Segments', self._menuViewSegments)
        self.vqAddMenuField('&View.&Symboliks', self._menuViewSymboliks)
        self.vqAddMenuField('&View.&Layouts.&Set Default', self._menuViewLayoutsSetDefault)
        self.vqAddMenuField('&View.&Layouts.&Save', self._menuViewLayoutsSave)
        self.vqAddMenuField('&View.&Layouts.&Load', self._menuViewLayoutsLoad)
        self.vqAddMenuField('&View.&Layouts.Load &Base Default', self._menuViewLayoutsLoadBase)

        self.vqAddMenuField('&Share.Share Workspace', self._menuShareWorkspace)
        self.vqAddMenuField('&Share.Connect to Shared Workspace', self._menuShareConnect)
        self.vqAddMenuField('&Share.Connect To Workspace Server', self._menuShareConnectServer)

        self.vqAddMenuField('&Tools.&Python', self._menuToolsPython)
        self.vqAddMenuField('&Tools.&Debug', self._menuToolsDebug)
        self.vqAddMenuField('&Tools.&Analysis.&Reanalyze Switchcase', self._menuToolsReSwitchCase)
        self.vqAddMenuField('&Tools.&Structures.Add Namespace', self._menuToolsStructNames)
        self.vqAddMenuField('&Tools.&Structures.New', self._menuToolsUStructNew)
        self.vqAddDynMenu('&Tools.&Structures.&Edit', self._menuToolsUStructEdit)

        self.vqAddDynMenu('&Tools.&Va Sets', self._menuToolsVaSets)

        self.vqAddMenuField('&Window.&Fullscreen', self._menuWindowFullscreen)
        self.vqAddMenuField('&Window.&Maximized', self._menuWindowMaximize)
        self.vqAddMenuField('&Window.&Normal', self._menuWindowNormal)

        self.vw.vprint('Welcome to Vivisect (Qt Edition)!')
        self.vw.vprint('Random Tip: %s' % viv_q_tips.getRandomTip())

        if len(self.vqGetDockWidgets()) == 0:
            self.vw.vprint('\n')
            #self.vw.vprint('&#10;')
            self.vw.vprint('Looks like you have an empty layout!')
            self.vw.vprint('Use View->Layouts->Load Base Default')

        fname = os.path.basename(self.vw.getMeta('StorageName', 'Unknown'))
        self.setWindowTitle('Vivisect: %s' % fname)
        self.windowState = QtCore.Qt.WindowNoState

        self.addHotKey('ctrl+o', 'file:open')
        self.addHotKeyTarget('file:open', self._menuFileOpen)
        self.addHotKey('ctrl+s', 'file:save')
        self.addHotKeyTarget('file:save', self._menuFileSave)
        self.addHotKey('ctrl+S', 'file:saveas')
        self.addHotKeyTarget('file:saveas', self._menuFileSaveAs)
        self.addHotKey('ctrl+meta+C', 'file:connecttoserver')
        self.addHotKeyTarget('file:connecttoserver', self._menuShareConnectServer)
        self.addHotKey('ctrl+w', 'file:quit')
        self.addHotKeyTarget('file:quit', self.close)
        self.addHotKey('ctrl+m', 'view:memory')
        self.addHotKey('ctrl+g', 'view:funcgraph')

    def vprint(self, msg, addnl=True):
        # ripped and modded from envi/cli.py
        self.vw.canvas.write(msg)
        if addnl:
            self.vw.canvas.write('\n')

    def getLocation(self, va):
        loctup = self.vw.getLocation(va)

        if loctup is None:
            self.vw.vprint('Location not found!')
        else:
            name = loc_type_names.get(loctup[L_LTYPE], 'Unspecified')
            self.vw.vprint('\nVA: %s' % hex(loctup[L_VA]))
            self.vw.vprint('    Size: %d' % loctup[L_SIZE])
            self.vw.vprint('    Type: %s' % name)
            self.vw.vprint('    Info: %s' % str(loctup[L_TINFO]))
            self.vw.vprint('    Repr: %s' % self.vw.reprLocation(loctup)[:64])

    def setVaName(self, va, parent=None):
        if parent is None:
            parent = self

        curname = self.vw.getName(va)
        if curname is None:
            curname = 'loc_%.8x' % va

        name, ok = QInputDialog.getText(parent, 'Enter...', 'Name', text=curname)
        if ok:
            name = str(name)
            if self.vw.vaByName(name):
                raise Exception('Duplicate Name: %s' % name)

            self.vw.makeName(va, name)

    def setName(self, va, tag, parent=None):
        '''
        Set the name for a given Tag or VA 
        '''
        if parent is None:
            parent = self

        if tag and tag[0] == 'name':
            # do tag things
            ttype, tagname = tag

            if tagname:
                tagname = tagname.decode('utf8')
                fva = self.vw.getFunction(va)
                if fva:
                    rtype, rname, cconv, cname, cargs = self.vw.getFunctionApi(fva)
                    if cargs:
                        for i, (atype, aname) in enumerate(cargs):
                            if aname == tagname:
                                self.setFuncArgName(fva, i, atype, aname)
                            else:
                                logger.warning("%s != %s" % (aname, tagname))
                    else:
                        logger.warning("setName(va=0x%x, tag=%r) called on a 'name' but function has no args: fva: 0x%x", va, repr(tag), fva)
                else:
                    vw.vprint("setName(va=0x%x, tag=%r):  can't determine what function we're in", va, repr(tag))
        else:
            self.setVaName(va)

    def setVaComment(self, va, parent=None):
        if parent is None:
            parent = self

        curcomment = self.vw.getComment(va)
        if curcomment is None:
            curcomment = ''

        comment, ok = QInputDialog.getText(parent, 'Enter...', 'Comment', text=curcomment)
        if ok:
            self.vw.setComment(va, str(comment))

    def setVaMultilineComment(self, va, parent=None):
        if parent is None:
            parent = self

        curcomment = self.vw.getComment(va)
        if curcomment is None:
            curcomment = ''

        comment, ok = QInputDialog.getMultiLineText(parent, 'Enter...', 'Comment', text=curcomment)
        if ok:
            self.vw.setComment(va, str(comment))

    def addVaXref(self, va, parent=None):
        if parent is None:
            parent = self
        xtova, ok = QInputDialog.getText(parent, 'Enter...', 'Make Code Xref 0x%x -> ' % va)
        if ok:
            try:
                val = self.vw.parseExpression(str(xtova))
                if self.vw.isValidPointer(val):
                    self.vw.addXref(va, val, REF_CODE)
                else:
                    self.vw.vprint("Invalid Expression: %s   (%s)" % (xtova, val))
            except Exception as e:
                self.vw.vprint(repr(e))

    def reanalyzeSwitchCase(self, va, parent=None):
        if parent is None:
            parent = self
        timestr, ok = QInputDialog.getText(parent, 'Re-Analyze Switchcase', 'Enter Timeout (secs) for analysis (0x%x): ' % va, text="300")
        if ok:
            try:
                timeout = self.vw.parseExpression(str(timestr))

                symswitch.analyzeJmp(self.vw, va, timeout=timeout)
            except Exception as e:
                self.vw.vprint(repr(e))

    def setFuncLocalName(self, fva, offset, atype, aname):
        curname = ''
        if self.vw.getFunctionLocal(fva, offset):
            curtype, curname = self.vw.getFunctionLocal(fva, offset)
        newname, ok = QInputDialog.getText(self, 'Enter...', 'Local Name', text=curname)
        if ok:
            self.vw.setFunctionLocal(fva, offset, LSYM_NAME, (atype, str(newname)))

    def setFuncArgName(self, fva, idx, atype, aname):
        curname = ''
        if len(self.vw.getFunctionArgs(fva)) > idx:
            curtype, curname = self.vw.getFunctionArgs(fva)[idx]
        newname, ok = QInputDialog.getText(self, 'Enter...', 'Argument Name', text=curname)
        if ok:
            self.vw.setFunctionArg(fva, idx, atype, str(newname))

    def showFuncCallGraph(self, fva):
        callview = viv_q_funcviews.FuncCallsView(self.vw, parent=self)
        callview.functionSelected(fva)
        callview.show()
        self.vqDockWidget(callview, floating=False)

    def makeStruct(self, va, parent=None):
        if parent is None:
            parent = self
        sname = vs_qt.selectStructure(self.vw.vsbuilder, parent=parent)
        if sname is not None:
            self.vw.makeStructure(va, sname)
        return sname

    def makePtrArray(self, va, break_on_bad=True, parent=None):
        if parent is None:
            parent = self
        # starting address
        # address after last element
        # count (alternate)
        ok, results = selectArrayBounds(va, va, 0, parent=parent)

        if ok:
            startva, stopva, count = results
            self.vw.makePointerArray(startva, stopva, count, break_on_bad)
            

    def addBookmark(self, va, parent=None):
        if parent is None:
            parent = self
        bname, ok = QInputDialog.getText(parent, 'Enter...', 'Bookmark Name')
        if ok:
            self.vw.setVaSetRow('Bookmarks', (va, str(bname)))

    def getMemoryWidgets(self):
        return self.views.get('VQVivMemoryView', [])

    def getMemWidgetsByName(self, name='viv', firstonly=True):
        '''
        Returns a list of Memory View Widgets with the given name.
        If "firstonly" is True, only return the first one or None(not a list)

        Returns a tuple of (Widget, DockWidget).  The "Widget" is obtained from
        the DockWidget, but they both have different powers.
        '''
        logger.debug("getWindowsByName(%r, firstonly=%r)", name, firstonly)
        out = []

        for vqDW in self.getMemoryWidgets():
            w = vqDW.widget()
            if w.getEnviNavName() == name:
                if firstonly:
                    return w, vqDW

                out.append((w,vqDW))

        if firstonly:   # if firstonly and we don't have one, return None
            return None

        return out

    def getFuncGraphs(self):
        return self.views.get('VQVivFuncgraphView', [])

    def getFuncGraphsByName(self, name='FuncGraph0', firstonly=True):
        '''
        Returns a list of Dock Widgets which have a "getEnviNavName"
        This includes MemoryViews and FuncGraphs
        '''
        logger.debug("getFuncGraphsByName()")

        out = []
        for vqDW in self.getFuncGraphs():
            w = vqDW.widget()
            if name != w.getEnviNavName():
                continue

            if firstonly:
                return w, vqDW

            out.append((w, vqDW))

        if firstonly:   # if firstonly and we don't have one, return None
            return None

        return out

    def sendMemWidgetTo(self, va, wname='viv', firstonly=False):
        '''
        Tells the named Envi Nav Widget to navigate to the given VA
        '''
        logger.debug("sendMemWidgetsTo(0x%x, wname=%r)", va, wname)
        for win in self.getMemWidgetsByName(wname, firstonly=False):
            w, vqFW = win

            logger.debug("sending %r to %r", w, hex(va))
            w.enviNavGoto(hex(va))
            if firstonly:
                break
        return True

    def sendFuncGraphTo(self, va, wname='FuncGraph0', firstonly=False):
        '''
        Tells the named Envi Nav Widget to navigate to the given VA
        '''
        logger.debug("sendFuncGraphTo(0x%x, wname=%r)", va, wname)
        for win in self.getFuncGraphsByName(wname, firstonly=False):
            w, vqFW = win

            logger.debug("sending %r to %r", w, hex(va))
            w.enviNavGoto(hex(va))
            if firstonly:
                break
        return True

    def getCliBar(self):
        '''
        Returns the CLI Bar object
        '''
        for c in self.children():
            if isinstance(c, vq_cli.VQCli):
                return c

    def getCliText(self):
        '''
        Get the text from the GUI's CLI Bar (at the bottom)
        '''
        cli = self.getCliBar()
        return cli.input.text()

    def setCliText(self, text):
        '''
        Set the text in the GUI's CLI Bar (at the bottom)
        '''
        logger.debug("setCliText(%r)" % text)
        cli = self.getCliBar()
        cli.input.setText(text)

    def _menuEditPrefs(self):
        configs = []
        configs.append(('Vivisect', self.vw.config.viv))
        configs.append(('Vdb', self.vw.config.vdb))
        self._cfg_widget = e_q_config.EnviConfigTabs(configs)
        self._cfg_widget.show()

    def _menuToolsUStructNew(self):
        u = viv_q_ustruct.UserStructEditor(self.vw)
        w = self.vqDockWidget(u, floating=True)
        w.resize(600, 600)

    def _menuToolsUStructEdit(self, name=None):
        if name is None:
            return self.vw.getUserStructNames()
        u = viv_q_ustruct.UserStructEditor(self.vw, name=name)
        w = self.vqDockWidget(u, floating=True)
        w.resize(600, 600)

    def _menuToolsVaSets(self, name=None):
        if name is None:
            return self.vw.getVaSetNames()
        view = viv_q_views.VQVivVaSetView(self.vw, self, name)
        self.vqDockWidget(view)

    def delFunction(self, fva, parent=None):
        if parent is None:
            parent = self

        yn, ok = QInputDialog.getItem(self, 'Delete Function', 'Confirm:', ('No', 'Yes'), 0, False)
        if ok and yn == 'Yes':
            self.vw.delFunction(fva)

    def vqInitDockWidgetClasses(self):

        exprloc = e_expr.MemoryExpressionLocals(self.vw, symobj=self.vw)
        exprloc['vw'] = self.vw
        exprloc['vwqgui'] = self
        exprloc['vprint'] = self.vw.vprint

        self.vqAddDockWidgetClass(viv_q_views.VQVivExportsView, args=(self.vw, self))
        self.vqAddDockWidgetClass(viv_q_views.VQVivFunctionsView, args=(self.vw, self))
        self.vqAddDockWidgetClass(viv_q_views.VQVivNamesView, args=(self.vw, self))
        self.vqAddDockWidgetClass(viv_q_views.VQVivImportsView, args=(self.vw, self))
        self.vqAddDockWidgetClass(viv_q_views.VQVivSegmentsView, args=(self.vw, self))
        self.vqAddDockWidgetClass(viv_q_views.VQVivStringsView, args=(self.vw, self))
        self.vqAddDockWidgetClass(viv_q_views.VQVivStructsView, args=(self.vw, self))
        self.vqAddDockWidgetClass(vq_python.VQPythonView, args=(exprloc, self))
        self.vqAddDockWidgetClass(viv_q_memory.VQVivMemoryView, args=(self.vw, self))
        self.vqAddDockWidgetClass(viv_q_funcgraph.VQVivFuncgraphView, args=(self.vw, self))
        self.vqAddDockWidgetClass(viv_q_symboliks.VivSymbolikFuncPane, args=(self.vw, self))

    @idlethread
    def vqRestoreGuiSettings(self, settings, guid=None):
        '''
        Restores GUI settings (size/layout/views) based on:
         * GUID
         * Filename(s)
         * Default Layout

        If workspace is connected to a server, we wait for a GUID to be present before proceeding
        '''

        if self.vw.server and not guid:
            # wait until the GUID has been loaded from the remote workspace before continuing
            self.vw._load_guid.wait() 

        if not guid:
            guid = self.vw.getVivGuid()

        logger.debug("vqRestoreGuiSettings() -> guid=%r  vw.server=%r", guid, self.vw.server)

        logger.debug("attempting to load GUI settings based on GUID: %s", guid)
        dwcls = settings.value('%s/DockClasses' % guid)
        state = settings.value('%s/DockState' % guid)
        geom = settings.value('%s/DockGeometry' % guid)
        stub = '%s/' % guid

        if compat_isNone(dwcls):
            names = list(self.vw.filemeta.keys())
            names.sort()
            name = '+'.join(names)
            logger.debug("attempting to load GUI settings based on Filename(s): %r", name)
            dwcls = settings.value('%s/DockClasses' % name)
            state = settings.value('%s/DockState' % name)
            geom = settings.value('%s/DockGeometry' % name)
            stub = '%s/' % name

        if compat_isNone(dwcls):
            logger.debug("loading default GUI settings")
            dwcls = settings.value('DockClasses')
            state = settings.value('DockState')
            geom = settings.value('DockGeometry')
            stub = ''

        if not compat_isNone(dwcls):
            for i, clsname in enumerate(dwcls):
                name = 'VQDockWidget%d' % i
                try:
                    tup = self.vqBuildDockWidget(str(clsname), floating=False)
                    if tup is not None:
                        d, obj = tup
                        d.setObjectName(name)
                        d.vqRestoreState(settings, name, stub)
                        d.show()
                except Exception as e:
                    self.vw.vprint('Error Building: %s: %s' % (clsname, e))

        # Once dock widgets are loaded, we can restoreState
        if not compat_isNone(state):
            self.restoreState(state)

        if not compat_isNone(geom):
            self.restoreGeometry(geom)

        # Just get all the resize activities done...
        vq_main.eatevents()
        for w in self.vqGetDockWidgets():
            w.show()

        return True

    def vqSaveGuiSettings(self, settings):

        dock_classes = []
        guid = self.vw.getVivGuid()
        names = list(self.vw.filemeta.keys())
        names.sort()
        vivname = '+'.join(names)

        # Enumerate the current dock windows and set
        # their names by their list order...
        for i, w in enumerate(self.vqGetDockWidgets()):
            widget = w.widget()
            dock_classes.append(widget.__class__.__name__)
            name = 'VQDockWidget%d' % i
            w.setObjectName(name)
            w.vqSaveState(settings, '%s/%s' % (guid, name))
            w.vqSaveState(settings, '%s/%s' % (vivname, name))

        geom = self.saveGeometry()
        state = self.saveState()

        # first store for this specific workspace
        settings.setValue('%s/DockClasses' % guid, dock_classes)
        settings.setValue('%s/DockGeometry' % guid, geom)
        settings.setValue('%s/DockState' % guid, state)

        # next store for this filename
        settings.setValue('%s/DockClasses' % vivname, dock_classes)
        settings.setValue('%s/DockGeometry' % vivname, geom)
        settings.setValue('%s/DockState' % vivname, state)
        # don't store the default.  that should be saved manually

    def vqGetDockWidgetsByName(self, name='viv', firstonly=False):
        '''
        Get list of DockWidgets matching a given name (default is 'viv').

        Returns a list of tuples (window, DockWidget)
        If firstonly==True, returns the first tuple, not a list.
        '''
        out = []
        for vqDW in self.vqGetDockWidgets():
            w = vqDW.widget()
            if hasattr(w, 'getEnviNavName') and w.getEnviNavName() == name:
                if firstonly:
                    return w, vqDW
                out.append((w,vqDW))
        return out

    def _menuToolsDebug(self):
        viv_vdbext.runVdb(self)

    def _menuFileOpen(self):
        # TODO: Add something to change the workspace storage name,
        # and also to list the currently loaded files
        # Right now it'll successively create storage files
        fname = getOpenFileName(self, 'Open...')
        if fname is None or not len(fname):
            return
        self.vw.vprint('Opening %s' % fname)
        self.setWindowTitle('Vivisect: %s' % fname)
        self.vw.loadFromFile(str(fname))
        self.vw.vprint('Analyzing %s' % fname)
        self.vw.analyze()
        self.vw.vprint('%s is ready!' % fname)

    @vq_main.workthread
    def _menuFileSave(self, fullsave=False, filename=None):
        if self.vw.server and filename is None:
            self.vw.vprint("Connected to remote workspace, not saving locally.")
            self.vw.vprint("Use 'File->Save As' to create a local backup copy of the workspace.")
            return

        # duplicate filename, since saveWorkspace() with None as the filename
        # forces a local save
        fname = filename
        if not fname:
            fname = self.vw.getMeta("StorageName")
        self.vw.vprint('Saving workspace... (%r)' % fname)

        try:
            self.vw.saveWorkspace(fullsave=fullsave, filename=filename)
        except Exception as e:
            self.vw.vprint(str(e))
        else:
            self.vw.vprint('complete!')

    def _menuFileSaveAs(self):
        fname = getSaveFileName(self, 'Save As...')
        if fname is None or not len(fname):
            return

        self._menuFileSave(fullsave=True, filename=fname)

    def _menuFileSaveServer(self):
        viv_q_remote.saveToServer(self.vw, parent=self)

    def _menuViewLayoutsLoadBase(self):
        dirname = os.path.dirname(viv_q_views.__file__) 
        fname = os.sep.join([dirname, "default.lyt"])
        settings = QtCore.QSettings(fname, QtCore.QSettings.IniFormat)
        self.vqRestoreGuiSettings(settings)

    def _menuViewLayoutsLoad(self):
        fname = getOpenFileName(self, 'Load Layout')
        if fname is None:
            return

        settings = QtCore.QSettings(fname, QtCore.QSettings.IniFormat)
        self.vqRestoreGuiSettings(settings)

    def _menuViewLayoutsSave(self):
        fname = getSaveFileName(self, 'Save Layout')
        if fname is None or not len(fname):
            return

        settings = QtCore.QSettings(fname, QtCore.QSettings.IniFormat)
        self.vqSaveGuiSettings(settings)

    def _menuViewLayoutsSetDefault(self):
        vq_app.VQMainCmdWindow.vqSaveGuiSettings(self, self._vq_settings)

    def _menuToolsReSwitchCase(self):
        timeoutSwitches = self.vw.getVaSetRows('SwitchCases_TimedOut')

        dynd = DynamicDialog('Reanalze Switchcase')
        dynd.addComboBox('select', title='Select Switchcase Branch', itemlist=\
                ['0x%x: (failed at %d secs)' % (va, tosec) for va, tosec in timeoutSwitches], \
                dfltidx=0)
        dynd.addIntHexField('timeout', title='Timeout (some analysis can be very long)', dflt=300)
        results = dynd.prompt()
        if not results:
            return

        timeout = results.get('timeout')
        vastr, _ = results.get('select').split(':',1)
        va = int(vastr, 16)
        
        symswitch.analyzeJmp(self.vw, va, timeout=timeout)

    def _menuToolsStructNames(self):
        nsinfo = vs_qt.selectStructNamespace()
        if nsinfo is not None:
            nsname, modname = nsinfo
            self.vw.vprint('Adding struct namespace: %s' % nsname)
            self.vw.addStructureModule(nsname, modname)

    def _menuShareWorkspace(self):
        self.vw.vprint('Sharing workspace...')
        daemon = viv_share.shareWorkspace(self.vw)
        self.vw.vprint('Workspace Listening Port: %d' % daemon.port)
        self.vw.vprint('Clients may now connect to your host on port %d' % daemon.port)

    def _menuShareConnect(self):
        viv_q_remote.openSharedWorkspace(self.vw, parent=self)

    def _menuShareConnectServer(self):
        viv_q_remote.openServerAndWorkspace(self.vw, parent=self)

    def _menuToolsPython(self):
        self.vqBuildDockWidget('VQPythonView', area=QtCore.Qt.RightDockWidgetArea)

    def _menuViewStrings(self):
        self.newStringsView()
    def _menuViewStructs(self):
        self.newStructsView()
    def _menuViewSegments(self):
        self.newSegmentsView()
    def _menuViewImports(self):
        self.newImportsView()
    def _menuViewExports(self):
        self.newExportsView()
    def _menuViewFunctions(self):
        self.newFunctionsView()
    def _menuViewNames(self):
        self.newNamesView()

    @firethread
    @vq_hotkey.hotkey('view:memory')
    def _menuViewMemory(self):
        self.newMemoryView()

    @firethread
    @vq_hotkey.hotkey('view:funcgraph')
    def _menuViewFuncGraph(self):
        self.newFuncGraphView()

    def _menuViewSymboliks(self):
        self.newSymbolikFuncView()

    @idlethread
    def newPythonView(self, floating=False):
        self.vqBuildDockWidget('VQPythonView', floating=floating, area=QtCore.Qt.RightDockWidgetArea)

    @idlethread
    def newStringsView(self, floating=False):
        self.vqBuildDockWidget('VQVivStringsView', floating=floating, area=QtCore.Qt.RightDockWidgetArea)

    @idlethread
    def newStructsView(self, floating=False):
        self.vqBuildDockWidget('VQVivStructsView', floating=floating, area=QtCore.Qt.RightDockWidgetArea)

    @idlethread
    def newSegmentsView(self, floating=False):
        self.vqBuildDockWidget('VQVivSegmentsView', floating=floating, area=QtCore.Qt.RightDockWidgetArea)

    @idlethread
    def newImportsView(self, floating=False):
        self.vqBuildDockWidget('VQVivImportsView', floating=floating, area=QtCore.Qt.RightDockWidgetArea)

    @idlethread
    def newExportsView(self, floating=False):
        self.vqBuildDockWidget('VQVivExportsView', floating=floating, area=QtCore.Qt.RightDockWidgetArea)

    @idlethread
    def newFunctionsView(self, floating=False):
        self.vqBuildDockWidget('VQVivFunctionsView', floating=floating, area=QtCore.Qt.RightDockWidgetArea)

    @idlethread
    def newNamesView(self, floating=False):
        self.vqBuildDockWidget('VQVivNamesView', floating=floating, area=QtCore.Qt.RightDockWidgetArea)

    @idlethread
    def newMemoryView(self, name='viv', floating=False, expr=None):
        dock, widget = self.vqBuildDockWidget('VQVivMemoryView', floating=floating, area=QtCore.Qt.TopDockWidgetArea)
        widget.setMemWindowName(name)
        if expr is None:
            expr = self._getFirstFileBase()
        widget.enviNavGoto(expr)

    def _getFirstFileBase(self):
        '''
        Returns a string expression of the first file registered in the workspace.
        If the filename is '' (a possibility), the ImageBase is returned from 
        file-metadata.
        '''
        files = self.vw.getFiles()
        if not len(files):
            return

        file = files[0]
        if not len(file):
            # a file may have a '' name
            imagebase = self.vw.getFileMeta(file, "ImageBase")
            if self.vw.isValidPointer(imagebase):
                return hex(imagebase)

        return file


    @idlethread
    def newFuncGraphView(self, name=None, floating=False, expr=None):
        dock, widget = self.vqBuildDockWidget('VQVivFuncgraphView', floating=floating, area=QtCore.Qt.TopDockWidgetArea)
        if name is not None:
            widget.setMemWindowName(name)
        if expr is not None:
            widget.enviNavGoto(expr)

    @idlethread
    def newSymbolikFuncView(self, floating=False):
        self.vqBuildDockWidget('VivSymbolikFuncPane', floating=floating, area=QtCore.Qt.TopDockWidgetArea)


    def _menuWindowFullscreen(self):
        if not self.windowState & QtCore.Qt.WindowFullScreen:
            self.windowState = QtCore.Qt.WindowFullScreen
            self.showFullScreen()
        else:
            self._menuWindowNormal()

    def _menuWindowMaximize(self):
        if not self.windowState & QtCore.Qt.WindowMaximized:
            self.windowState = QtCore.Qt.WindowMaximized
            self.showMaximized()

    def _menuWindowNormal(self):
        if not self.windowState & QtCore.Qt.WindowNoState:
            self.windowState = QtCore.Qt.WindowNoState
            self.showNormal()

    @vq_main.idlethread
    def _ve_fireEvent(self, event, edata):
        return viv_base.VivEventDist._ve_fireEvent(self, event, edata)

def selectArrayBounds(startva, stopva, count=0, parent=None):
    if not stopva:
        stopva = startva

    dynd = DynamicDialog('Pointer Array Dialog', parent=parent)

    try:
        dynd.addIntHexField('startva', dflt=hex(startva), title='Start Address')
        dynd.addIntHexField('stopva', dflt=hex(stopva), title='Stop Address ')
        dynd.addIntHexField('count', dflt=0, title='Count (if nonzero, ignore stop address)')
    except Exception:
        logger.warning("ERROR BUILDING DIALOG!", exc_info=1)

    results = dynd.prompt()

    ok =  len(results) != 0
    if ok:
        startva = results.get('startva')
        stopva = results.get('stopva')
        count = results.get('count')
        return ok, (startva, stopva, count)
    return False, None

@vq_main.idlethread
def runqt(vw, closeme=None):
    '''
    Use this API to instantiate a QT main window and show it when
    there is already a main thread running...
    '''
    mw = VQVivMainWindow(vw)
    viv_extensions.loadExtensions(vw, mw)
    mw.show()

    if closeme:
        closeme.close()

    return mw

def main(vw):
    vq_main.startup(css=vq_colors.qt_matrix)
    mw = VQVivMainWindow(vw)
    viv_extensions.loadExtensions(vw, mw)
    mw.show()
    vq_main.main()


if __name__ == '__main__':
    vw = viv_cli.VivCli()
    if len(sys.argv) == 2:
        vw.loadWorkspace(sys.argv[1])
    main(vw)



# Mnemonic distribution
# Reverse depth

columns = (
    ("Code Blocks", int),
    ("Mnem Dist", int),
)


def report(vw):
    ret = {}
    cbtot = {}
    for f in vw.getFunctions():
        fblocks = vw.getFunctionBlocks(f)
        cbtot[f] = len(fblocks)

    for f, c in cbtot.items():
        mndist = vw.getFunctionMeta(f, "MnemDist", -1)
        ret[f] = (c, mndist)

    return ret



import logging

from enum import IntEnum

import vstruct

from vstruct.primitives import *

logger = logging.getLogger(__name__)

class UnicodeString(vstruct.VStruct):
    def __init__(self):
        vstruct.VStruct.__init__(self)
        self.length = v_uint32()  # size represents the number of 16-bit unicode characters

    def pcb_length(self):
        if self.length > 0:
            numbytes = 2 * self.length
            self.valu = v_wstr(size=numbytes)
            numpad = ((numbytes + 3) & ~0x3) - numbytes
            self.padding = v_bytes(size=numpad)

class FileTime(vstruct.VStruct):
    def __init__(self):
        vstruct.VStruct.__init__(self)
        self.dwLowDateTime = v_uint32()
        self.dwHighDateTime  = v_uint32()

class Blob(vstruct.VStruct):
    def __init__(self):
        vstruct.VStruct.__init__(self)
        self.size = v_uint32()

    def pcb_size(self):
        self.data = v_bytes(size=self.size)

class CodePageString(vstruct.VStruct):
    def __init__(self):
        vstruct.VStruct.__init__(self, unicode=False)
        self.size = v_uint32()
        if unicode:
            self.characters = v_wstr(size=2*self.size)
        else:
            self.characters = v_str(size=self.size)

class VersionedStream(vstruct.VStruct):
    def __init__(self):
        vstruct.VStruct.__init__(self)
        self.guid = GUID()
        self.name = CodePageString()

class ClipboardData(vstruct.VStruct):
    def __init__(self):
        vstruct.VStruct.__init__(self)

        self.size = v_uint32()
        self.format = v_uint32()

    def pcb_format(self):
        numbytes = (self.size + 3) & ~0x3
        self.data = v_bytes(size=numbytes)

class Decimal(vstruct.VStruct):
    def __init__(self):
        vstruct.VStruct.__init__(self)
        self.reserved = v_uint16()
        self.scale = v_uint8()
        self.sign = v_uint8()
        self.high32 = v_uint32()
        self.low64 = v_uint64()

class ArrayDimension(vstruct.VStruct):
    def __init__(self):
        vstruct.VStruct.__init__(self)
        self.size = v_uint32()
        self.indexOff = v_uint32()

class ArrayHeader(vstruct.VStruct):
    def __init__(self):
        vstruct.VStruct.__init__(self)
        self.type = v_uint32()
        self.numDim = v_uint32()

        self.dimensions = vstruct.VArray()

    def pcb_numDim(self):
        for i in range(self.numDim):
            self.dimensions.vsAddElement(ArrayDimension())

class PropertyType(IntEnum):
    VT_EMPTY = 0x0000
    VT_NULL = 0x0001
    VT_I2 = 0x0002
    VT_I4 = 0x0003
    VT_R4 = 0x0004
    VT_R8 = 0x0005
    VT_CY = 0x0006
    VT_DATE = 0x0007
    VT_BSTR = 0x0008

    VT_ERROR = 0x000A
    VT_BOOL = 0x000B

    VT_DECIMAL = 0x000E
    VT_I1 = 0x0010
    VT_UI1 = 0x0011
    VT_UI2 = 0x0012
    VT_UI4 = 0x0013
    VT_I8 = 0x0014
    VT_UI8 = 0x0015
    VT_INT = 0x0016
    VT_UINT = 0x0017
    VT_LPSTR = 0x001E
    VT_LPWSTR = 0x001F

    VT_FILETIME = 0x0040
    VT_BLOB = 0x0041
    VT_STREAM = 0x0042
    VT_STORAGE = 0x0043
    VT_STREAMED_OBJECT = 0x0044
    VT_STORED_OBJECT = 0x0045
    VT_BLOB_OBJECT = 0x0046
    VT_CF = 0x0047
    VT_CLSID = 0x0048
    VT_VERSIONED_STREAM = 0x0049

    # These are special. They're modifiers to the above
    VT_VECTOR = 0x1000
    VT_ARRAY = 0x2000

    # VECTOR. See MS-OLEPS Section 2.15 for why there are gaps because this list
    # is *not* equivalent to just OR-ing the above items with VT_VECTOR
    VT_VECTOR_I2 = VT_VECTOR | VT_I2
    VT_VECTOR_I4 = VT_VECTOR | VT_I4
    VT_VECTOR_R4 = VT_VECTOR | VT_R4
    VT_VECTOR_R8 = VT_VECTOR | VT_R8
    VT_VECTOR_CY = VT_VECTOR | VT_CY
    VT_VECTOR_DATE = VT_VECTOR | VT_DATE
    VT_VECTOR_BSTR = VT_VECTOR | VT_BSTR
    VT_VECTOR_ERROR = VT_VECTOR | VT_ERROR
    VT_VECTOR_BOOL = VT_VECTOR | VT_BOOL

    VT_VECTOR_VARIANT = VT_VECTOR | 0x000C

    VT_VECTOR_I1 = VT_VECTOR | VT_I1
    VT_VECTOR_UI1 = VT_VECTOR | VT_UI1
    VT_VECTOR_UI2 = VT_VECTOR | VT_UI2
    VT_VECTOR_UI4 = VT_VECTOR | VT_UI4
    VT_VECTOR_I8 = VT_VECTOR | VT_I8
    VT_VECTOR_UI8 = VT_VECTOR | VT_UI8
    VT_VECTOR_LPSTR = VT_VECTOR | VT_LPSTR
    VT_VECTOR_LPWSTR = VT_VECTOR | VT_LPWSTR

    VT_VECTOR_FILETIME = VT_VECTOR | VT_FILETIME
    VT_VECTOR_CF = VT_VECTOR | VT_CF
    VT_VECTOR_CLSID = VT_VECTOR | VT_CLSID

    # Similar to VECTOR. There are gaps, so see Section 2.15
    VT_ARRAY_I2 = VT_ARRAY | VT_I2
    VT_ARRAY_I4 = VT_ARRAY | VT_I4
    VT_ARRAY_R4 = VT_ARRAY | VT_R4
    VT_ARRAY_R8 = VT_ARRAY | VT_R8
    VT_ARRAY_CY = VT_ARRAY | VT_CY
    VT_ARRAY_DATE = VT_ARRAY | VT_DATE
    VT_ARRAY_BSTR = VT_ARRAY | VT_BSTR
    VT_ARRAY_ERROR = VT_ARRAY | VT_ERROR
    VT_ARRAY_BOOL = VT_ARRAY | VT_BOOL

    VT_ARRAY_VARIANT = VT_ARRAY | 0x000C

    VT_ARRAY_DECIMAL = VT_ARRAY | VT_DECIMAL
    VT_ARRAY_I1 = VT_ARRAY | VT_I1
    VT_ARRAY_UI1 = VT_ARRAY | VT_UI1
    VT_ARRAY_UI2 = VT_ARRAY | VT_UI2
    VT_ARRAY_UI4 = VT_ARRAY | VT_UI4
    VT_ARRAY_INT = VT_ARRAY | VT_INT
    VT_ARRAY_UINT = VT_ARRAY | VT_UINT


class TypedPropertyValue(vstruct.VStruct):
    # See MS-OLEPS for the MS definintion of this (Specifically section 2.15)
    def __init__(self, unicode=False):
        vstruct.VStruct.__init__(self)

        self.typeid = v_uint16()
        self.padding = v_uint16()

        self.unicode = unicode

        # self.vectorheader = None
        # self.arrayheader = None
        # self.valu = None
        # self.pad = None

    def pcb_typeid(self):

        if self.typeid & PropertyType.VT_VECTOR:
            self.vectorheader = v_uint32()
        elif self.typeid & PropertyType.VT_ARRAY:
            logger.warning('ArrayHeaders are not yet fully supported')
            #self.arrayheader = ArrayHeader()
        else:
            valu, pad = self._getTypeValu(self.typeid)
            self.valu = valu
            if pad:
                self.pad = pad

    def pcb_vectorheader(self):
        self.valu = vstruct.VArray()
        for i in range(self.vectorheader):
            if self.typeid == PropertyType.VT_VECTOR_VARIANT:
                valu = TypedPropertyValue(unicode=self.unicode)
            else:
                valu, _ = self._getTypeValu(self.typeid)

            self.valu.vsAddElement(valu)

        # there's a couple vector ones we've gotta pad to get a total amount of space that's
        # a multiple of 4
        if self.typeid == PropertyType.VT_VECTOR_I2:
            size = 2
        elif self.typeid == PropertyType.VT_VECTOR_UI2:
            size = 2
        elif self.typeid == PropertyType.VT_VECTOR_BOOL:
            size = 2
        elif self.typeid == PropertyType.VT_VECTOR_I1:
            size = 1
        elif self.typeid == PropertyType.VT_VECTOR_UI1:
            size = 1
        else:
            return

        nxt = ((size * self.vectorheader) + 3) & ~0x3
        self.pad = v_bytes(size=nxt - (size * self.vectorheader))

    def _getTypeValu(self, typeid):
        valu = None
        pad = None
        match typeid:
            case PropertyType.VT_I2:
                valu = v_int16()
                pad = v_uint16()
            case PropertyType.VT_I4:
                valu = v_int32()
            case PropertyType.VT_R4:
                valu = v_float()
            case PropertyType.VT_R8:
                valu = v_double()
            case PropertyType.VT_CY:
                valu = v_uint64()
            case PropertyType.VT_DATE:
                valu = v_double()
            case PropertyType.VT_BSTR:
                valu = CodePageString(unicode=self.unicode)
            case PropertyType.VT_ERROR:
                valu = v_uint32()
            case PropertyType.VT_BOOL:
                valu = v_uint16()
            case PropertyType.VT_DECIMAL:
                valu = Decimal()
            case PropertyType.VT_I1:
                valu = v_int8()
            case PropertyType.VT_UI1:
                valu = v_uint8()
            case PropertyType.VT_UI2:
                valu = v_uint16()
            case PropertyType.VT_UI4:
                valu = v_uint32()
            case PropertyType.VT_I8:
                valu = v_int64()
            case PropertyType.VT_UI8:
                valu = v_uint64()
            case PropertyType.VT_INT:
                valu = v_int32()
            case PropertyType.VT_UINT:
                valu = v_uint32()
            case PropertyType.VT_LPSTR:
                valu = CodePageString(unicode=self.unicode)
            case PropertyType.VT_LPWSTR:
                valu = UnicodeString()
            case PropertyType.VT_FILETIME:
                valu = FileTime()
            case PropertyType.VT_BLOB:
                valu = Blob()
            case PropertyType.VT_STREAM:
                valu = CodePageString(unicode=self.unicode)
            case PropertyType.VT_STORAGE:
                valu = CodePageString(unicode=self.unicode)
            case PropertyType.VT_STREAMED_OBJECT:
                valu = CodePageString(unicode=self.unicode)
            case PropertyType.VT_STORED_OBJECT:
                valu = CodePageString(unicode=self.unicode)
            case PropertyType.VT_BLOB_OBJECT:
                valu = Blob()
            case PropertyType.VT_CF:
                valu = ClipboardData()
            case PropertyType.VT_CLSID:
                valu = GUID()
            case PropertyType.VT_VERSIONED_STREAM:
                valu = VersionedStream()

        return valu, pad

class StringPropertyValue(vstruct.VStruct):
    # All strings specified by this structure are unicode
    # See MS-PROPSTORE
    def __init__(self, unicode=False):
        vstruct.VStruct.__init__(self)
        self.unicode = unicode
        self.size = v_uint32()
        self.namesize = v_uint32()

    def pcb_namesize(self):
        if self.size > 0:
            self.reserved = v_uint8()

            self.name = v_zwstr()

            self.valu = TypedPropertyValue(unicode=self.unicode)

class IntegerPropertyValue(vstruct.VStruct):
    def __init__(self, unicode=False):
        vstruct.VStruct.__init__(self)
        self.unicode = unicode
        self.size = v_uint32()
        self.id = v_uint32()

    def pcb_id(self):
        if self.size > 0:
            self.reserved = v_uint8()

            self.valu = TypedPropertyValue(unicode=self.unicode)

STRING_PROP_GUID = '{D5CDD505-2E9C-101B-9397-08002B2CF9AE}'
# This is what MS calls it, but we could use a better name
class PropertyStorage(vstruct.VStruct):
    def __init__(self):
        vstruct.VStruct.__init__(self)
        self.size = v_uint32()

    def pcb_size(self):
        # A size of 0 indicates we're done
        if self.size != 0:
            self.version = v_uint32()
            self.formatid = GUID()
            self.propvalues = vstruct.VArray()

    def vsParse(self, bytes, offset=0):
        # parse the base stuff and check if we've got anything
        retoff = vstruct.VStruct.vsParse(self, bytes, offset=offset)
        if self.size == 0:
            return len(self) + offset

        ctor = StringPropertyValue if self.formatid == STRING_PROP_GUID else IntegerPropertyValue

        # parse until a propvalu comes back with a *value size* of 0
        step = len(self)
        unicode = False
        while True:
            elem = None
            if self.formatid == STRING_PROP_GUID:
                elem = StringPropertyValue()
                elem.vsParse(bytes[(offset+step):])
            else:
                elem = IntegerPropertyValue()
                elem.vsParse(bytes[(offset+step):])

                if elem.id == 0x1 and (elem.valu.valu == 1200 or elem.valu.valu == 1201):
                    unicode = True

            self.propvalues.vsAddElement(elem)
            step += elem.size
            if elem.size == 0:
                break

        return len(self) + offset



"""
envi.bytesig signatures for Microsoft Visual Studio
"""
import binascii

import envi.bytesig as e_bytesig

sigs = [
    ('680000000064a10000000050', 'ff00000000ffffffffffffff', 'ntdll.seh3_prolog'),
    ('8b4df064890d00000000595f5e5bc951c3', None, 'ntdll.seh3_epilog'),
    ('680000000064ff35000000008b442410', 'ff00000000ffffffffffffffffffffff', 'ntdll.seh4_prolog'),
    # Seen in 32-bit samples using VS 2005, 2008, 2010, 2012, 2013.
    ('8b4df064890d00000000595f5f5e5b8be55d51c3', None, 'ntdll.seh4_epilog'),
    # Seen in 32-bit samples using VS 2015, 2017 15.0.
    ('8b4df064890d00000000595f5f5e5b8be55d51f2c3', None, 'ntdll.seh4_epilog'),
    ('a10000000033c58945fc', 'ff00000000ffffffffff', 'ntdll.gs_prolog'),

    # Seen in 32-bit samples using VS 6.
    ('513d001000008d4c2408721481e9001000002d0010000085013d0010000073ec2bc88bc485018be18b088b400450c3', None, 'ntdll._alloca_probe'),
    # Seen in 32-bit samples using VS 2005, 2008, 2010, 2012, 2013.
    ('518d4c24042bc81bc0f7d023c88bc42500f0ffff3bc8720a8bc159948b00890424c32d001000008500ebe9', None, 'ntdll._alloca_probe'),
    # Seen in 32-bit samples using VS 2015, 2017, 2019.
    ('518d4c24042bc81bc0f7d023c88bc42500f0ffff3bc8f2720b8bc159948b00890424f2c32d001000008500ebe7', None, 'ntdll._alloca_probe'),
    # Seen in 64-bit samples using VS 2005, 2008, 2010, 2012, 2013, 2015.
    ('4883ec104c8914244c895c24084d33db4c8d5424184c2bd04d0f42d3654c8b1c25100000004d3bd37316664181e200f04d8d9b00f0ffff41c603004d3bd375f04c8b14244c8b5c24084883c410c3', None, 'ntdll._alloca_probe'),
    # Seen in 64-bit samples using VS 2017, 2019.
    ('4883ec104c8914244c895c24084d33db4c8d5424184c2bd04d0f42d3654c8b1c25100000004d3bd3f27317664181e200f04d8d9b00f0ffff41c603004d3bd3f275ef4c8b14244c8b5c24084883c410f2c3', None, 'ntdll._alloca_probe'),

#    ('3b0d000000000f85afdc0200c3', 'ffff00000000ffffffffffffff', 'ntdll.security_check_cookie'),
    # 32-bit assembly used in VS 2005, 2008, 2010, 2012, 2013.
    ('3b0d000000007502f3c3e9', 'ffff00000000ffffffffff','ntdll.security_check_cookie'),
    # 32-bit assembly used in VS 2015, 2017 (includes bnd opcodes).
    ('3b0d00000000f27502f2c3f2e9', 'ffff00000000ffffffffffffff', 'ntdll.security_check_cookie'),
    # 64-bit assembly used in VS 2005, 2008, 2010, 2012, 2013.
    ('483b0d00000000751148c1c11066f7c1ffff7502f3c348c1c910e9', 'ffffff00000000ffffffffffffffffffffffffffffffffffffffff', 'ntdll.security_check_cookie_64'),
    # 64-bit assembly used in VS 2015.
    ('483b0d00000000f2751148c1c11066f7c1fffff27502f2c348c1c910e9', 'ffffff00000000ffffffffffffffffffffffffffffffffffffffffffff', 'ntdll.security_check_cookie_64'),
    # 64-bit assembly used in VS 2019.
    ('483b0d00000000f2751248c1c11066f7c1fffff27502f2c348c1c910e9', 'ffffff00000000ffffffffffffffffffffffffffffffffffffffffffff', 'ntdll.security_check_cookie_64'),

    ('8bff558bec5151a300000000890d00000000891500000000891d00000000893500000000893d000000008c15000000008c0d000000008c1d000000008c05000000008c25000000008c2d000000009c',
     'ffffffffffffffff00000000ffff00000000ffff00000000ffff00000000ffff00000000ffff00000000ffff00000000ffff00000000ffff00000000ffff00000000ffff00000000ffff00000000ff',
     'ntdll.report_gsfailure',
    ),
    ('8bff558bec81ec28030000a300000000890d00000000891500000000891d00000000893500000000893d00000000668c1500000000668c0d00000000668c1d00000000668c0500000000668c2500000000668c2d000000009c',
     'ffffffffffffffffffffffff00000000ffff00000000ffff00000000ffff00000000ffff00000000ffff00000000ffffff00000000ffffff00000000ffffff00000000ffffff00000000ffffff00000000ffffff00000000ff',
     'ntdll.report_gsfailure',
    ),

    ('6aff5064a100000000508b44240c64892500000000896c240c8d6c240c50c3',None, 'ntdll.eh_prolog'),

]

class VisualStudioVamp(e_bytesig.SignatureTree):

    def __init__(self):
        e_bytesig.SignatureTree.__init__(self)
        for bytez, masks, fname in sigs:
            bytez = binascii.unhexlify(bytez)
            if masks is not None:
                masks = binascii.unhexlify(masks)
            self.addSignature(bytez, masks=masks, val=fname)

# seh3_prolog
#.text:0x7c8024d6  68c09a837c       push kernel32.seh3_handler
#.text:0x7c8024db  64a100000000     fs: mov eax,dword [0x00000000]
#.text:0x7c8024e1  50               push eax
#.text:0x7c8024e2  8b442410         mov eax,dword [esp + 16]
#.text:0x7c8024e6  896c2410         mov dword [esp + 16],ebp
#.text:0x7c8024ea  8d6c2410         lea ebp,dword [esp + 16]
#.text:0x7c8024ee  2be0             sub esp,eax
#.text:0x7c8024f0  53               push ebx
#.text:0x7c8024f1  56               push esi
#.text:0x7c8024f2  57               push edi
#.text:0x7c8024f3  8b45f8           mov eax,dword [ebp - 8]
#.text:0x7c8024f6  8965e8           mov dword [ebp - 24],esp
#.text:0x7c8024f9  50               push eax
#.text:0x7c8024fa  8b45fc           mov eax,dword [ebp - 4]
#.text:0x7c8024fd  c745fcffffffff   mov dword [ebp - 4],4294967295
#.text:0x7c802504  8945f8           mov dword [ebp - 8],eax
#.text:0x7c802507  8d45f0           lea eax,dword [ebp - 16]
#.text:0x7c80250a  64a300000000     fs: mov dword [0x00000000],eax
#.text:0x7c802510  c3               ret 

# seh3_epilog
#.text:0x7c802511  8b4df0           mov ecx,dword [ebp - 16]
#.text:0x7c802514  64890d00000000   fs: mov dword [0x00000000],ecx
#.text:0x7c80251b  59               pop ecx
#.text:0x7c80251c  5f               pop edi
#.text:0x7c80251d  5e               pop esi
#.text:0x7c80251e  5b               pop ebx
#.text:0x7c80251f  c9               leave 
#.text:0x7c802520  51               push ecx
#.text:0x7c802521  c3               ret 


# seh4_prolog
#.text:0x00402390  68f0234000       push hello.seh4_handler
#.text:0x00402395  64ff3500000000   fs: push dword [0x00000000]
#.text:0x0040239c  8b442410         mov eax,dword [esp + 16]
#.text:0x004023a0  896c2410         mov dword [esp + 16],ebp
#.text:0x004023a4  8d6c2410         lea ebp,dword [esp + 16]
#.text:0x004023a8  2be0             sub esp,eax
#.text:0x004023aa  53               push ebx
#.text:0x004023ab  56               push esi
#.text:0x004023ac  57               push edi
#.text:0x004023ad  a110c44000       mov eax,dword [loc_0040c410]
#.text:0x004023b2  3145fc           xor dword [ebp - 4],eax
#.text:0x004023b5  33c5             xor eax,ebp
#.text:0x004023b7  50               push eax
#.text:0x004023b8  8965e8           mov dword [ebp - 24],esp
#.text:0x004023bb  ff75f8           push dword [ebp - 8]
#.text:0x004023be  8b45fc           mov eax,dword [ebp - 4]
#.text:0x004023c1  c745fcfeffffff   mov dword [ebp - 4],4294967294
#.text:0x004023c8  8945f8           mov dword [ebp - 8],eax
#.text:0x004023cb  8d45f0           lea eax,dword [ebp - 16]
#.text:0x004023ce  64a300000000     fs: mov dword [0x00000000],eax
#.text:0x004023d4  c3               ret 

# seh4_epilog
#.text:0x004025e9  8b4df0           mov ecx,dword [ebp - 16]
#.text:0x004025ec  64890d00000000   fs: mov dword [0x00000000],ecx
#.text:0x004025f3  59               pop ecx
#.text:0x004025f4  5f               pop edi
#.text:0x004025f5  5f               pop edi
#.text:0x004025f6  5e               pop esi
#.text:0x004025f7  5b               pop ebx
#.text:0x004025f8  8be5             mov esp,ebp
#.text:0x004025fa  5d               pop ebp
#.text:0x004025fb  51               push ecx
#.text:0x004025fc  c3               ret 

# gs_prolog_chunk
#.text:0x00409362  a11cc04000       mov eax,dword [GS_COOKIE]
#.text:0x00409367  33c5             xor eax,ebp
#.text:0x00409369  8945fc           mov dword [ebp - 4],eax

# GS PROLOG CAN JUST TAKE LOCALS SIZE!

# GS Epilog (is ecx ok chunk)
#.text:0x0040119a  3b0d1cc04000     cmp ecx,dword [GS_COOKIE]
#.text:0x004011a0  7502             jnz loc_004011a4
#.text:0x004011a2  f3c3             rep: ret 
#.text:0x004011a4  loc_004011a4: [1 XREFS]
#.text:0x004011a4  e903160000       jmp loc_004027ac

# Also  __report_gs_failure is loc_004027ac at this point

# EH Prolog:
#6aff             push 255
#50               push eax                          ; Create the new record
#64a100000000     fs: mov eax,dword [0x00000000]
#50               push eax                          ; Grab and save off the next
#8b44240c         mov eax,dword [esp + local_1]     ; Saved EIP -> eax
#64892500000000   fs: mov dword [0x00000000],esp    ; ptr to 3 dword record into FS:0
#896c240c         mov dword [esp + local_1],ebp     ; Put old EBP into saved EIP spot (esp now delta 16 total)
#8d6c240c         lea ebp,dword [esp + local_1]     ; Make ebp point to itself
#50               push eax
#c3               ret 





from PyQt5.QtWidgets import QVBoxLayout

import vdb.qt.base
import envi.qt.memwrite as e_qt_mw

from vqt.main import *

class VdbMemWriteWindow(vdb.qt.base.VdbWidgetWindow):

    def __init__(self, db, dbt, expr='', esize='', parent=None):
        vdb.qt.base.VdbWidgetWindow.__init__(self, db, dbt, parent=parent)

        self.memWriteWidget = e_qt_mw.MemWriteWindow(expr=expr, esize=esize, emu=dbt, parent=parent)
        vbox = QVBoxLayout()
        vbox.addWidget(self.memWriteWidget)
        self.setLayout(vbox)

        self.memWriteWidget.writeToMemory.connect(self.writeMemory)
        self.setWindowTitle('Write Memory')

        vqtconnect(self.vdbWriteMem, 'vdb:writemem')

    def vdbWriteMem(self, event, einfo):
        self.vqLoad()

    def vqLoad(self):
        self.memWriteWidget.renderMemory()

    def enviNavGoto(self, expr, esize='', rend=''):
        self.memWriteWidget.setValues(expr, esize)
        self.vqLoad()

    def vqGetSaveState(self):
        expr, esize = self.memWriteWidget.getValues()
        return {'expr': expr, 'esize': esize}

    def vqSetSaveState(self, state):
        self.memWriteWidget.setValues(state.get('expr', ''), state.get('esize', ''))

    def writeMemory(self, expr, hexbytez):
        self.db.do_writemem('-X %s %s' % (expr, hexbytez))



import envi.archs.arm as e_arm

class ArmMixin(e_arm.ArmModule, e_arm.ArmRegisterContext):

    def __init__(self):
        e_arm.ArmModule.__init__(self)
        e_arm.ArmRegisterContext.__init__(self)

    def archGetStackTrace(self):
        pc = self.getProgramCounter()
        sp = self.getStackCounter()
        return [ (pc,sp), ]




import envi
import envi.bits as e_bits
import envi.bytesig as e_bsig

from envi.archs.z80.regs import *
from envi.archs.z80.const import *

import envi.archs.z80.z80opcode as z80_opcode

sigtree = e_bsig.SignatureTree()

for row in z80_opcode.z80table:
    sighex, maskhex = row[0]

    sig = binascii.unhexlify(sighex)
    mask = binascii.unhexlify(maskhex)

    sigtree.addSignature(sig, masks=mask, val=row)


class z80RegOper(envi.RegisterOper):
    def __init__(self, reg):
        self.reg = reg

class z80ImmOper(envi.ImmedOper):
    def __init__(self, imm):
        self.imm = imm

    def repr(self, op):
        return '%.4xH' % self.imm

class z80ConstOper(z80ImmOper):
    pass

class z80RegMem(envi.DerefOper):
    def __init__(self, reg, disp = 0):
        self.reg = reg
        self.disp = disp

    def repr(self, op):
        rname = regctx.getRegisterName(self.reg)
        if self.disp > 0:
            return '(%s + %d)' % (rname, self.disp)
        if self.disp < 0:
            return '(%s - %d)' % (rname, abs(self.disp))
        return '(%s)' % rname

class z80Opcode(envi.Opcode):
    pass

class z80Disasm:

    def __init__(self):
        # NOTE: For eventual things like "modes" etc...
        pass

    def disasm(self, bytez, offset, va):
        row = sigtree.getSignature(bytez, offset)
        if row is None:
            raise envi.InvalidInstruction(bytez=bytez[offset:offset+8], va=va)
        sigmask, mnem, o1type, o1info, o2type, o2info, oplen, immoff, iflags = row
        #ret = i386Opcode(va, optype, mnem, prefixes, (offset-startoff)+operoffset, operands, iflags)
        opers = []
        if o1type is not None:
            opers.append(self._buildOper(bytez, offset, immoff, o1type, o1info))
        if o2type is not None:
            opers.append(self._buildOper(bytez, offset, immoff, o2type, o2info))
        return z80Opcode(va, 0, mnem, 0, oplen, opers, iflags)

    def _buildOper(self, bytez, offset, immoff, otype, oinfo):

        if otype == OPTYPE_Reg:
            return z80RegOper(oinfo)

        elif otype == OPTYPE_RegMem:
            return z80RegMem(oinfo)

        elif otype == OPTYPE_const:
            return z80ConstOper(oinfo)

        elif otype == OPTYPE_imm8:
            imm = e_bits.parsebytes(bytez, offset+immoff, 1)
            return z80ImmOper(imm)

        elif otype == OPTYPE_imm16:
            imm = e_bits.parsebytes(bytez, offset+immoff, 2)
            return z80ImmOper(imm)

        elif otype == OPTYPE_RegAlt:
            return z80RegOper(oinfo)

        elif otype == OPTYPE_Ind:
            pass

        elif otype == OPTYPE_RegMemDisp:
            disp = e_bits.parsebytes(bytez, offset+immoff, 1, sign=True)
            return z80RegMem(oinfo, disp)

        else:
            raise Exception('Unknown z80 operand type: %d' % otype)



import vtrace.qt
import envi.common as e_common

import envi.qt.memory
import envi.qt.memcanvas

from vqt.main import *
from vqt.common import *

from PyQt5 import QtCore
from PyQt5.QtWidgets import QApplication

class VDBACT:

    def __init__(self, db, cmdline):
        self.db = db
        self.cmdline = cmdline

    def __call__(self, *args, **kwargs):
        workthread(self.db.onecmd)(self.cmdline)

class VdbMemoryCanvas(envi.qt.memcanvas.VQMemoryCanvas):
    # We get self.db set by our memory window smash...
    def initMemWindowMenu(self, va, menu):

        t = self.db.getTrace()

        menu.addAction('Run To Here', VDBACT(self.db, 'go 0x%.8x' % va))
        menu.addAction('Add Breakpoint', VDBACT(self.db, 'bp 0x%.8x' % va))

        bp = t.getBreakpointByAddr(va)
        if bp is not None:
            bpid = bp.getId()
            menu.addAction('Remove Breakpoint', VDBACT(self.db, 'bp -r %d' % bpid))

            if bp.enabled:
                menu.addAction('Disable Breakpoint', VDBACT(self.db, 'bp -d %d' % bpid))
            else:
                menu.addAction('Enable Breakpoint', VDBACT(self.db, 'bp -e %d' % bpid))

        smenu_patch = menu.addMenu('Bytes')
        nop = t.archGetNopInstr()
        # yuck...need to fix this on gui re-architecture.
        currend = self.vdb_memwin.rend_select.currentText()
        if nop is not None and 'asm' in currend:
            smenu_patch.addAction('Set Bytes To NOP', ACT(self._menuSetOpTo, va, nop))

        if 'asm' in currend:
            smenu_patch.addAction('Set Bytes To NULL', ACT(self._menuSetOpTo, va, '\x00'))

        if 'asm' in currend or 'bytes' in currend:
            smenu_patch.addAction('Modify Bytes', ACT(self._menuWriteMem, va))

        smenu_patch.addAction('Copy Bytes To Clipboard', ACT(self._menuCopyBytesToClipBoard, va, currend, False))
        smenu_patch.addAction('Copy Bytes To Clipboard (All Window Bytes)', ACT(self._menuCopyBytesToClipBoard, va, currend, True))

        return envi.qt.memcanvas.VQMemoryCanvas.initMemWindowMenu(self, va, menu)

    def menuLeftClick(self):
        '''
        handles left clicking on the 'open in current renderer'.
        '''
        self._menuFollow(self._canv_curva)

    def mouseDoubleClickEvent(self, event):
        if event.button() != QtCore.Qt.LeftButton:
            return

        self._menuFollow(self._canv_curva)
        clipboard = QApplication.clipboard()
        clipboard.setText(hex(self._canv_curva))

    def _menuSetOpTo(self, va, byte):
        # TODO: this should be a helper func somewhere, nopInstruction?
        t = self.db.getTrace()
        op = t.parseOpcode(va)
        oplen = len(op)
        nlen = len(byte)

        if oplen % nlen != 0:
            raise Exception('cannot nop all bytes, mismatch')

        bytez = byte * (oplen / nlen)
        self.db.do_writemem('%s %s' % (va, bytez))

    def _menuCopyBytesToClipBoard(self, va, currend, all_window_bytes):
        t = self.db.getTrace()
        if all_window_bytes:
            va = self._canv_beginva
            size = self._canv_endva - va

        else:
            if 'asm' in currend:
                op = t.parseOpcode(va)
                size = len(op)
            else:
                size = t.getPointerSize()

        bytez = t.readMemory(va, size)

        clipboard = QApplication.clipboard()
        clipboard.setText(e_common.hexify(bytez))

    def _menuFollow(self, va, rend='', newWindow=False):
        totalsize = self._canv_endva - self._canv_beginva

        if newWindow:
            vqtevent('vdb:view:memory', (hex(va), '256', rend))
            return

        if self._canv_navcallback:
            self._canv_navcallback(hex(va), str(totalsize), rend)

    def _menuWriteMem(self, va):
        vqtevent('vdb:view:writemem', (hex(va), '256'))

    def reRender(self):
        '''
        Forces the canvas to refresh everything it's currently displaying.
        '''
        if self._canv_endva is None or self._canv_beginva is None:
            # not rendered yet.
            return

        totalsize = self._canv_endva - self._canv_beginva

        self.renderMemory(self._canv_beginva, totalsize)

class VdbMemoryWindow(envi.qt.memory.VQMemoryWindow, vtrace.qt.VQTraceNotifier):

    def __init__(self, db, dbt, parent=None, expr=None, sizeexpr=None, rend=None, **kwargs):
        vtrace.qt.VQTraceNotifier.__init__(self, trace=dbt)
        envi.qt.memory.VQMemoryWindow.__init__(self, dbt, syms=dbt, parent=parent, **kwargs)

        for rname in db.canvas.getRendererNames():
            self.mem_canvas.addRenderer(rname, db.canvas.getRenderer(rname))

        self.mem_canvas.db = db
        self.mem_canvas.vdb_memwin = self

        self.loadRendSelect()

        if not dbt.isAttached():
            self.setEnabled(False)
        elif dbt.isRunning():
            self.setEnabled(False)
        else:
            self._renderMemory()

        vqtconnect(self._renderMemory, 'vdb:stopped')
        vqtconnect(self._renderMemory, 'vdb:writemem')
        vqtconnect(self._renderMemory, 'vdb:delbreak')
        vqtconnect(self._renderMemory, 'vdb:addbreak')

        vqtconnect(ACT(self.setEnabled, True), 'vdb:attached')
        vqtconnect(ACT(self.setEnabled, False), 'vdb:detached')

    def initMemoryCanvas(self, memobj, syms=None):
        return VdbMemoryCanvas(memobj, syms=syms, parent=self)

    def vqLoad(self):
        self._renderMemory()



'''
Structure definitions for the OSX MachO binary format.
'''
import struct
import logging
import binascii

import vstruct
import envi.bits as e_bits
from vstruct.defs.macho.fat import *
from vstruct.defs.macho.const import *
from vstruct.defs.macho.stabs import *
from vstruct.defs.macho.loader import *

logger = logging.getLogger(__name__)

class mach_o(vstruct.VStruct):

    def __init__(self):
        vstruct.VStruct.__init__(self)
        self._raw_bytes = b''
        self._symbols = None
        self._entrypoints = None
        self._struct_info = []

        self.mach_header = mach_header()
        self.load_commands = vstruct.VStruct()

    def getPointerSize(self):
        return self.mach_header.vsGetMeta('psize')

    def getSymbols(self):
        if self._symbols is not None:
            return self._symbols

        self._symbols = []
        for fname, vs in self.load_commands:
            if vs.cmd != LC_SYMTAB:
                continue
            strbytes = self._raw_bytes[vs.stroff:vs.stroff+vs.strsize]
            strtab = strbytes.split('\x00')
            offset = vs.symoff
            for i in range(vs.nsyms):
                n = nlist() # FIXME 64!
                offset = n.vsParse(self._raw_bytes, offset)
                #symstr = strtab[n.n_strx]
                # FIXME this is slow!
                symstr = strbytes[n.n_strx:].split('\x00', 1)[0]

    def getEntryPoints(self):
        if self._entrypoints is not None:
            return self._entrypoints

        self._entrypoints = []
        offset = len(self.mach_header)
        for fname, vs in self.load_commands:
            if vs.cmd in (LC_MAIN, LC_UNIXTHREAD):
                eoff = len(vs)
                psize = vs.flavor
                bigend = self.vsGetEndian()

                fmt = e_bits.getFormat(psize, bigend)

                for x in range(vs.count):
                    ptr, = struct.unpack_from(fmt, self._raw_bytes, eoff+offset)
                    self._entrypoints.append((offset + eoff, ptr))
                    eoff += psize

            offset += vs.cmdsize

        return self._entrypoints

    def getLibDeps(self):
        '''
        Return a list of the library files this Mach-O is dependant on
        '''
        ret = []
        for fname, vs in self.load_commands:
            if vs.cmd != LC_LOAD_DYLIB:
                continue
            ret.append(vs.namedata)
        return ret

    def getSegments(self):
        '''
        Return a list of (segname, rva, perms, bytes) tuples for the memory
        segments defined by the loader commands
        '''
        ret = []
        for fname, vs in self.load_commands:
            if vs.cmd not in (LC_SEGMENT, LC_SEGMENT_64):
                continue
            if vs.segname == '__PAGEZERO':
                continue

            # Slice the segment bytes from raw bytes
            fbytes = self._raw_bytes[ vs.fileoff: vs.fileoff + vs.filesize ]
            # Pad out to virtual size
            try:
                fbytes = fbytes.ljust(vs.vmsize, b'\x00')
            except:
                logger.warning('Segment allocation failure')
                continue

            ret.append((vs.segname, vs.vmaddr, vs.initprot, fbytes))
        return ret

    def getStructureInfo(self):
        '''
        Returns Structure information usable by the loader to tag file structures
        '''
        structs = []
        cmdsyms = []
        for fname, vs in self.load_commands:
            if vs.cmd not in (LC_SEGMENT, LC_SEGMENT_64):
                continue
            if vs.segname == '__PAGEZERO':
                continue

            if vs.fileoff == 0:
                offset = 0
                va = None
                if self.mach_header.vsGetMeta('psize') == 8:
                    va = vs.vmaddr
                    structs.append(('macho.mach_header_64', va))
                    offset += len(mach_header_64())

                else:
                    va = vs.vmaddr
                    structs.append(('macho.mach_header', va))
                    offset += len(mach_header())

                for fcmd, fvs in self.load_commands:
                    structs.append(('macho.' + fvs.vsGetTypeName(), va + offset))
                    offset += fvs.cmdsize
                    cmdsyms.append((va+offset, fcmd))

        return structs, cmdsyms

    def vsParse(self, bytes, offset=0):
        '''
        Parses entire MachO structure-set
        '''
        magic = struct.unpack('<I', bytes[:4])[0]
        if magic in (MH_MAGIC_64, MH_CIGAM_64):
            self.mach_header = mach_header_64()

        self._raw_bytes = bytes[offset:]
        offset = self.mach_header.vsParse(bytes, offset=offset)

        # no callback is fired for mach_header, set we set it here
        self.vsSetEndian(self.mach_header.vsGetEndian())
        bigend = self.vsGetEndian()
        fmt = ('<II','>II')[bigend]

        for i in range(self.mach_header.ncmds):
            cmdtype, cmdlen = struct.unpack(fmt, bytes[offset:offset+8])
            cmdclass = getCommandClass(cmdtype)
            cmdobj = cmdclass(bigend=bigend)
            cmdobj.vsParse(bytes, offset=offset)
            self.load_commands.vsAddField('cmd%d' % i, cmdobj)
            offset += cmdobj.cmdsize



import os
import logging

from PyQt5 import QtCore
from PyQt5.QtWidgets import *

import vqt.cli as vq_cli
import vqt.main as vq_main
import vqt.saveable as vq_save
import vqt.hotkeys as vq_hotkeys
import vqt.menubuilder as vq_menu
from vqt.saveable import compat_isNone

logger = logging.getLogger(__name__)

class VQDockWidget(vq_hotkeys.HotKeyMixin, QDockWidget):

    def __init__(self, parent):
        QDockWidget.__init__(self, parent)
        vq_hotkeys.HotKeyMixin.__init__(self)
        self.addHotKey('ctrl+enter', 'mem:undockmaximize')
        self.addHotKeyTarget('mem:undockmaximize', self._hotkey_undock_maximize)
        self.setAllowedAreas(QtCore.Qt.AllDockWidgetAreas)

    def vqSaveState(self, settings, name, stub=''):
        wid = self.widget()
        if isinstance(wid, vq_save.SaveableWidget):
            return wid.vqSaveState(settings, name, stub)

    def vqRestoreState(self, settings, name, stub=''):
        wid = self.widget()
        if isinstance(wid, vq_save.SaveableWidget):
            return wid.vqRestoreState(settings, name, stub)

    def setWidget(self, widget):
        # If he sets his window title, we want to...
        self.setWindowTitle(widget.windowTitle())
        widget.setWindowTitle = self.setWindowTitle
        QDockWidget.setWidget(self, widget)

    def closeEvent(self, event):

        self.hide()
        parent = self.parent()

        w = self.widget()

        # remove widget from the views dict
        clsname = w.__class__.__name__
        dwidgets = parent.views.get(clsname)
        if self in dwidgets:
            dwidgets.remove(self)

        w.setParent(None)
        w.close()

        # remove widget from main _vq_dockwidgets list
        parent.vqRemoveDockWidget(self)

        event.accept()

    def _hotkey_undock_maximize(self):
        # if docked, undock
        if not self.isFloating():
            self.setFloating(1)
            # if not maximized, maximize
            if not self.isMaximized():
                self.showMaximized()
            else:
                # else dock
                self.showNormal()
                self.setFloating(False)
        else:
            # else dock
            self.showNormal()
            self.setFloating(False)

        self.show()
        self.raise_()

class VQMainCmdWindow(vq_hotkeys.HotKeyMixin, QMainWindow):
    '''
    A base class for application window's to inherit from.
    '''

    __cli_widget_class__ = vq_cli.VQCli

    def __init__(self, appname, cmd, **kwargs):

        super(QMainWindow, self).__init__(**kwargs)
        vq_hotkeys.HotKeyMixin.__init__(self)

        self._vq_appname = appname
        self._vq_dockwidgets = []
        self.views = {}

        self._vq_settings = QtCore.QSettings('invisigoth', application=appname, parent=self)
        self._vq_histfile = os.path.join(os.path.expanduser('~'), '.%s_history' % appname)

        self._dock_classes = {}

        self.vqInitDockWidgetClasses()

        self._vq_mbar = vq_menu.VQMenuBar()
        self.setMenuBar(self._vq_mbar)

        # AnimatedDocks, AllowNestedDocks, AllowTabbedDocks, ForceTabbedDocks, VerticalTabs
        self.setDockOptions(self.AnimatedDocks | self.AllowTabbedDocks)

        self._vq_cli = self.__cli_widget_class__(cmd)
        self._vq_cli.input.loadHistory(self._vq_histfile)
        self._vq_cli.sigCliQuit.connect( self.close )

        self.setCentralWidget(self._vq_cli)
        self.vqRestoreGuiSettings(self._vq_settings)

    def vqAddMenuField(self, fname, callback, args=()):
        self._vq_mbar.addField(fname, callback, args=args)

    def vqAddDynMenu(self, fname, callback):
        self._vq_mbar.addDynMenu(fname, callback)

    def vqInitDockWidgetClasses(self):
        # apps can over-ride
        pass

    def vqAddDockWidgetClass(self, cls, args=()):
        self._dock_classes[cls.__name__] = (cls, args)

    def vqBuildDockWidget(self, clsname, floating=False, area=QtCore.Qt.TopDockWidgetArea):
        res = self._dock_classes.get(clsname)
        if res is None:
            logger.warning('vqBuildDockWidget Failed For: %s (No class constructor found)', clsname)
            return
        cls, args = res
        obj = cls(*args)
        return (self.vqDockWidget(obj, area, floating=floating), obj)

    def vqRestoreGuiSettings(self, settings, stub=''):
        dwcls = settings.value('DockClasses')

        if not compat_isNone(dwcls):
            for i, clsname in enumerate(dwcls):
                name = 'VQDockWidget%d' % i
                try:
                    # we haven't loaded the extensions yet, so all we'll have is the base clases
                    if str(clsname) not in self._dock_classes:
                        continue
                    tup = self.vqBuildDockWidget(str(clsname), floating=False)
                    if tup is not None:
                        d, obj = tup
                        d.setObjectName(name)
                        d.vqRestoreState(settings, name, stub)
                        d.show()
                except Exception as e:
                    logger.error('Error Building: %s: %s', clsname, e)

        # Once dock widgets are loaded, we can restoreState
        state = settings.value('DockState')
        if not compat_isNone(state):
            self.restoreState(state)

        geom = settings.value('DockGeometry')
        if not compat_isNone(geom):
            self.restoreGeometry(geom)

        # Just get all the resize activities done...
        vq_main.eatevents()
        for w in self.vqGetDockWidgets():
            w.show()

        return True

    def vqSaveGuiSettings(self, settings, stub=''):

        dock_classes = []

        # Enumerate the current dock windows and set
        # their names by their list order...
        for i, w in enumerate(self.vqGetDockWidgets()):
            widget = w.widget()
            dock_classes.append(widget.__class__.__name__)
            name = 'VQDockWidget%d' % i
            w.setObjectName(name)
            w.vqSaveState(settings,name,stub)

        settings.setValue('DockClasses', dock_classes)
        settings.setValue('DockGeometry', self.saveGeometry())
        settings.setValue('DockState', self.saveState())

    def closeEvent(self, event):
        self.vqSaveGuiSettings(self._vq_settings)
        self._vq_cli.input.saveHistory(self._vq_histfile)
        QMainWindow.closeEvent(self, event)

    def vqGetDockWidgets(self):
        return list(self._vq_dockwidgets)

    def vqClearDockWidgets(self):
        for wid in self.vqGetDockWidgets():
            wid.close()

    def vqRemoveDockWidget(self, widget):
        self._vq_dockwidgets.remove(widget)
        self.removeDockWidget(widget)

    def vqDockWidget(self, widget, area=QtCore.Qt.TopDockWidgetArea, floating=False):
        d = VQDockWidget(self)
        d.setWidget(widget)
        d.setFloating(floating)
        self.addDockWidget(area, d)
        self._vq_dockwidgets.append(d)

        # store the widget by classname
        clsname = widget.__class__.__name__
        dwidgets = self.views.get(clsname)
        if dwidgets is None:
            dwidgets = []
            self.views[clsname] = dwidgets
        dwidgets.append(d)

        self.restoreDockWidget(d)
        d.show()
        return d
